# EpyNN/epynn/network/models.py
# Standard library imports
import traceback
import time

# Related third party imports
import numpy as np

# Local application/library specific imports
from epynn.commons.io import (
    encode_dataset,
    scale_features,
)
from epynn.commons.library import write_model
from epynn.commons.loss import loss_functions
from epynn.commons.models import dataSet
from epynn.network.report import (
    model_report,
    initialize_model_report,
    single_batch_report,
)
from epynn.commons.plot import pyplot_metrics
from epynn.network.initialize import (
    model_assign_seeds,
    model_initialize,
    model_initialize_exceptions,
)
from epynn.network.hyperparameters import (
    model_hyperparameters,
    model_learning_rate,
)
from epynn.network.evaluate import model_evaluate
from epynn.network.forward import model_forward
from epynn.network.backward import model_backward
from epynn.network.training import model_training
from epynn.settings import se_hPars


class EpyNN:
    """
    Definition of a Neural Network prototype following the EpyNN scheme.

    :param layers: Network architecture.
    :type layers: list[Object]

    :param name: Name of network, defaults to 'EpyNN'.
    :type name: str, optional
    """

    def __init__(self, layers, name='EpyNN'):
        """Initialize instance variable attributes.

        :ivar layers: Network architecture.
        :vartype layers: list[Object]

        :ivar embedding: Embedding layer.
        :vartype embedding: :class:`epynn.embedding.models.Embedding`

        :ivar ts: Timestamp identifier.
        :vartype ts: int

        :ivar uname: Network unique identifier.
        :vartype uname: str

        :ivar initialized: Model initialization state.
        :vartype initialized: bool
        """
        # Layers
        self.layers = layers
        self.embedding = self.layers[0]

        # Identification
        self.ts = int(time.time())
        self.uname = str(self.ts) + '_' + name

        # State
        self.initialized = False

        return None

    def forward(self, X):
        """Wrapper for :func:`epynn.network.forward.model_forward()`.

        :param X: Set of sample features.
        :type X: :class:`numpy.ndarray`

        :return: Output of forward propagation through all layers in the Network.
        :rtype: :class:`numpy.ndarray`
        """
        A = model_forward(self, X)

        return A

    def backward(self, dA):
        """Wrapper for :func:`epynn.network.backward.model_backward()`.

        :param dA: Derivative of the loss function with respect to the output of forward propagation.
        :type dA: :class:`numpy.ndarray`
        """
        dX = model_backward(self, dA)

        return None

    def initialize(self, loss='MSE', se_hPars=se_hPars, metrics=['accuracy'], seed=None, params=True, end='\n'):
        """Wrapper for :func:`epynn.network.initialize.model_initialize()`. Perform a dry epoch including all but not the parameters update step.

        :param loss: Loss function to use for training, defaults to 'MSE'. See :py:mod:`epynn.commons.loss` for built-in functions.
        :type loss: str, optional

        :param se_hPars: Global hyperparameters, defaults to :class:`epynn.settings.se_hPars`. If local hyperparameters were assigned to one layer, these remain unchanged.
        :type se_hPars: dict[str: float or str], optional

        :param metrics: Metrics to monitor and print on terminal report or plot, defaults to ['accuracy']. See :py:mod:`epynn.commons.metrics` for built-in metrics. Note that it also accept loss functions string identifiers.
        :type metrics: list[str], optional

        :param seed: Reproducibility in pseudo-random procedures.
        :type seed: int or NoneType, optional

        :param params: Layer parameters initialization, defaults to `True`.
        :type params: bool, optional

        :param end: Whether to print every line for initialization steps or overwrite, default to `\\n`.
        :type end: str in ['\\n', '\\r'], optional
        """
        # Initialize model summary
        self.network = {id(layer):{} for layer in self.layers}

        # Initialize storage for selected metrics evaluation
        metrics = metrics.copy()
        metrics.append(loss)

        self.metrics = {m:[[] for _ in range(3)] for m in metrics}

        # Check consistency output activation and loss
        self.output = self.layers[-1].activation['activate']
        self.training_loss = loss_functions(loss, self.output)

        # Assign model and layers hyperparameters
        self.se_hPars = se_hPars
        model_hyperparameters(self)

        # Seed model and layers
        self.seed = seed
        model_assign_seeds(self)

        try:
            # Attempt to initialize model
            model_initialize(self, params=params, end=end)

        except Exception:
            # Handle errors and provide debug info
            trace = traceback.format_exc()
            model_initialize_exceptions(self, trace)

        # Termination
        self.initialized = True

        return None

    def train(self, epochs, verbose=None, init_logs=True):
        """Wrapper for :func:`epynn.network.training.model_training()`. Apart, it computes learning rate along learning epochs.

        :param epochs: Number of training iterations.
        :type epochs: int

        :param verbose: Print logs every Nth epochs, defaults to `None` which sets to every tenth of epochs.
        :type verbose: int or NoneType, optional

        :param init_logs: Print data, architecture and hyperparameters logs, defaults to `True`.
        :type init_logs: bool, optional
        """
        # Model initialization
        if not self.initialized:
            self.initialize()

        # Handling training intiation or continuation scenarii
        self.epochs = epochs if not self.e else epochs + self.e + 1

        # Compute learning rate schedule for layers in model
        model_learning_rate(self)

        if init_logs:
            # From model.initialize() method
            initialize_model_report(self, timeout=3)

        if not verbose:
            # By defaut, store full evaluation one every tenth of epochs
            verbose = epochs // 10 if epochs >= 10 else 1

        # Start training
        self.verbose = verbose
        self.cts = time.time()

        model_training(self)

        return None

    def evaluate(self):
        """Wrapper for :func:`epynn.network.evaluate.model_evaluate()`. Good spot for further implementation of early stopping procedures.
        """
        model_evaluate(self)

        return None

    def write(self, path=None):
        """Write model on disk.

        :param path: Path to write the model on disk, defaults to `None` which writes in the `models` subdirectory created from :func:`epynn.commons.library.configure_directory()`.
        :type path: str or NoneType, optional
        """
        write_model(self, path)

        return None

    def batch_report(self, batch, A):
        """Wrapper for :func:`epynn.network.report.single_batch_report()`.
        """
        single_batch_report(self, batch, A)

        return None

    def report(self):
        """Wrapper for :func:`epynn.network.report.model_report()`.
        """
        model_report(self)

        return None

    def plot(self, pyplot=True, path=None):
        """Wrapper for :func:`epynn.commons.plot.pyplot_metrics()`. Plot metrics from model training.

        :param pyplot: Plot of results on GUI using matplotlib.
        :type pyplot: bool, optional

        :param path: Write matplotlib plot, defaults to `None` which writes in the `plots` subdirectory created from :func:`epynn.commons.library.configure_directory()`. To not write the plot at all, set to `False`.
        :type path: str or bool or NoneType, optional
        """
        if pyplot:
            pyplot_metrics(self, path)

        return None

    def predict(self, X_data, X_encode=False, X_scale=False):
        """Perform prediction of label from unlabeled samples in dataset.

        :param X_data: Set of sample features.
        :type X_data: list[list[int or float or str]] or :class:`numpy.ndarray`

        :param X_encode: One-hot encode sample features, defaults to `False`.
        :type X_encode: bool, optional

        :param X_scale: Normalize sample features within [0, 1] along all axis, default to `False`.
        :type X_scale: bool, optional

        :return: Data embedding and output of forward propagation.
        :rtype: :class:`epynn.commons.models.dataSet`
        """
        X_data = np.array(X_data)

        if X_encode:
            # One-hot encoding using embedding layer cache
            element_to_idx = self.embedding.e2i
            elements_size = self.embedding.d['e']
            X_data = encode_dataset(X_data, element_to_idx, elements_size)

        if X_scale:
            # Array-wide normalization in [0, 1]
            X_data = scale_features(X_data)

        dset = dataSet(X_data)

        # Predict
        dset.A = self.forward(dset.X)

        # Check label encoding
        encoded = (self.embedding.dtrain.Y.shape[1] > 1)

        # Make decisions
        dset.P = np.argmax(dset.A, axis=1) if encoded else np.around(dset.A)

        return dset



# EpyNN/epynnlive/dummy_image/prepare_dataset.py
# Standard library imports
import random

# Related third party imports
import numpy as np


def features_image(WIDTH=28, HEIGHT=28):
    """Generate dummy image features.

    :param WIDTH: Image width, defaults to 28.
    :type WIDTH: int

    :param HEIGHT: Image height, defaults to 28.
    :type HEIGHT: int

    :return: Random image features of size N_FEATURES.
    :rtype: :class:`numpy.ndarray`

    :return: Non-random image features of size N_FEATURES.
    :rtype: :class:`numpy.ndarray`
    """
    # Number of channels is one for greyscale images
    DEPTH = 1

    # Number of features describing a sample
    N_FEATURES = WIDTH * HEIGHT * DEPTH

    # Number of distinct tones in features
    N_TONES = 16

    # Shades of grey
    GSCALE = [i for i in range(N_TONES)]

    # Random choice of shades for N_FEATURES iterations
    features = [random.choice(GSCALE) for j in range(N_FEATURES)]

    # Vectorization of features
    features = np.array(features).reshape(HEIGHT, WIDTH, DEPTH)

    # Masked features
    mask_on_features = features.copy()
    mask_on_features[np.random.randint(0, HEIGHT)] = np.zeros_like(features[0])
    mask_on_features[:, np.random.randint(0, WIDTH)] = np.zeros_like(features[:, 0])

    # Random choice between random image or masked image
    features = random.choice([features, mask_on_features])

    return features, mask_on_features


def label_features(features, mask_on_features):
    """Prepare label associated with features.

    The dummy law is:

    Image is NOT random = positive
    Image is random = negative

    :param features: Random image features of size N_FEATURES
    :type features: :class:`numpy.ndarray`

    :param mask_on_features: Non-random image features of size N_FEATURES
    :type mask_on_features: :class:`numpy.ndarray`

    :return: Single-digit label with respect to features
    :rtype: int
    """
    # Single-digit positive and negative labels
    p_label = 0
    n_label = 1

    # Test if image is not random (0)
    if np.sum(features) == np.sum(mask_on_features):
        label = p_label

    # Test if image is random (1)
    elif np.sum(features) != np.sum(mask_on_features):
        label = n_label

    return label


def prepare_dataset(N_SAMPLES=100):
    """Prepare a set of dummy time sample features and label.

    :param N_SAMPLES: Number of samples to generate, defaults to 100.
    :type N_SAMPLES: int

    :return: Set of sample features.
    :rtype: tuple[:class:`numpy.ndarray`]

    :return: Set of single-digit sample label.
    :rtype: tuple[int]
    """
    # Initialize X and Y datasets
    X_features = []
    Y_label = []

   # Iterate over N_SAMPLES
    for i in range(N_SAMPLES):

        # Compute random string features
        features, mask_on_features = features_image()

        # Retrieve label associated with features
        label = label_features(features, mask_on_features)

        # Append sample features to X_features
        X_features.append(features)

        # Append sample label to Y_label
        Y_label.append(label)

    # Prepare X-Y pairwise dataset
    dataset = list(zip(X_features, Y_label))

    # Shuffle dataset
    random.shuffle(dataset)

    # Separate X-Y pairs
    X_features, Y_label = zip(*dataset)

    return X_features, Y_label



# EpyNN/epynnlive/captcha_mnist/prepare_dataset.py
# Standard library imports
import tarfile
import random
import gzip
import os

# Related third party imports
import wget
import numpy as np

# Local application/library specific imports
from epynn.commons.logs import process_logs


def download_mnist():
    """Download a subset of the MNIST database.
    """
    data_path = os.path.join('.', 'data')

    if not os.path.exists(data_path):

        # Download @url with wget
        url = 'https://synthase.s3.us-west-2.amazonaws.com/mnist_database.tar'
        fname = wget.download(url)

        # Extract archive
        tar = tarfile.open(fname).extractall('.')
        process_logs('Make: '+fname, level=1)

        # Clean-up
        os.remove(fname)

    return None


def prepare_dataset(N_SAMPLES=100):
    """Prepare a dataset of hand-written digits as images.

    :param N_SAMPLES: Number of MNIST samples to retrieve, defaults to 100.
    :type N_SAMPLES: int

    :return: Set of sample features.
    :rtype: tuple[:class:`numpy.ndarray`]

    :return: Set of single-digit sample label.
    :rtype: tuple[:class:`numpy.ndarray`]
    """
    # Process MNIST images
    img_file = gzip.open('data/train-images-idx3-ubyte.gz')

    header = img_file.read(16)
    image_size = int.from_bytes(header[8:12], byteorder='big')
    buf = img_file.read(image_size * image_size * N_SAMPLES)
    X_features = np.frombuffer(buf, dtype=np.uint8).astype(np.float32)
    X_features = X_features.reshape(N_SAMPLES, image_size, image_size, 1)

    # Process MNIST labels
    label_file = gzip.open('data/train-labels-idx1-ubyte.gz')

    header = label_file.read(8)
    buf = label_file.read(image_size * image_size * N_SAMPLES)
    Y_label = np.frombuffer(buf, dtype=np.uint8)

    # Prepare X-Y pairwise dataset
    dataset = list(zip(X_features, Y_label))

    # Shuffle dataset
    random.shuffle(dataset)

    # Separate X-Y pairs
    X_features, Y_label = zip(*dataset)

    return X_features, Y_label



# EpyNN/epynn/gru/parameters.py
# Related third party imports
import numpy as np


def gru_compute_shapes(layer, A):
    """Compute forward shapes and dimensions from input for layer.
    """
    X = A    # Input of current layer

    layer.fs['X'] = X.shape    # (m, s, e)

    layer.d['m'] = layer.fs['X'][0]    # Number of samples (m)
    layer.d['s'] = layer.fs['X'][1]    # Steps in sequence (s)
    layer.d['e'] = layer.fs['X'][2]    # Elements per step (e)

    # Parameter Shapes             Unit cells (u)
    eu = (layer.d['e'], layer.d['u'])    # (e, u)
    uu = (layer.d['u'], layer.d['u'])    # (u, u)
    u1 = (1, layer.d['u'])               # (1, u)
    # Update gate    Reset gate       Hidden hat
    layer.fs['Uz'] = layer.fs['Ur'] = layer.fs['Uhh'] = eu
    layer.fs['Vz'] = layer.fs['Vr'] = layer.fs['Vhh'] = uu
    layer.fs['bz'] = layer.fs['br'] = layer.fs['bhh'] = u1

    # Shape of hidden state (h) with respect to steps (s)
    layer.fs['h'] = (layer.d['m'], layer.d['s'], layer.d['u'])

    return None


def gru_initialize_parameters(layer):
    """Initialize trainable parameters from shapes for layer.
    """
    # For linear activation of update gate (z_)
    layer.p['Uz'] = layer.initialization(layer.fs['Uz'], rng=layer.np_rng)
    layer.p['Vz'] = layer.initialization(layer.fs['Vz'], rng=layer.np_rng)
    layer.p['bz'] = np.zeros(layer.fs['bz']) # dot(X, U) + dot(hp, V) + b

    # For linear activation of reset gate (r_)
    layer.p['Ur'] = layer.initialization(layer.fs['Ur'], rng=layer.np_rng)
    layer.p['Vr'] = layer.initialization(layer.fs['Vr'], rng=layer.np_rng)
    layer.p['br'] = np.zeros(layer.fs['br']) # dot(X, U) + dot(hp, V) + b

    # For linear activation of hidden hat (hh_)
    layer.p['Uhh'] = layer.initialization(layer.fs['Uhh'], rng=layer.np_rng)
    layer.p['Vhh'] = layer.initialization(layer.fs['Vhh'], rng=layer.np_rng)
    layer.p['bhh'] = np.zeros(layer.fs['bhh']) # dot(X, U) + dot(r * hp, V) + b

    return None


def gru_compute_gradients(layer):
    """Compute gradients with respect to weight and bias for layer.
    """
    # Gradients initialization with respect to parameters
    for parameter in layer.p.keys():
        gradient = 'd' + parameter
        layer.g[gradient] = np.zeros_like(layer.p[parameter])

    # Reverse iteration over sequence steps
    for s in reversed(range(layer.d['s'])):

        X = layer.fc['X'][:, s]      # Input for current step
        hp = layer.fc['hp'][:, s]    # Previous hidden state

        # (1) Gradients of the loss with respect to U, V, b
        dhh_ = layer.bc['dhh_'][:, s]            # Gradient w.r.t hidden hat hh_
        layer.g['dUhh'] += np.dot(X.T, dhh_)     # (1.1) dL/dUhh
        layer.g['dVhh'] += np.dot((layer.fc['r'][:, s] * hp).T, dhh_)
        layer.g['dbhh'] += np.sum(dhh_, axis=0)  # (1.3) dL/dbhh

        # (2) Gradients of the loss with respect to U, V, b
        dz_ = layer.bc['dz_'][:, s]             # Gradient w.r.t update gate z_
        layer.g['dUz'] += np.dot(X.T, dz_)      # (2.1) dL/dUz
        layer.g['dVz'] += np.dot(hp.T, dz_)     # (2.2) dL/dVz
        layer.g['dbz'] += np.sum(dz_, axis=0)   # (2.3) dL/dbz

        # (3) Gradients of the loss with respect to U, V, b
        dr_ = layer.bc['dr_'][:, s]             # Gradient w.r.t reset gate r_
        layer.g['dUr'] += np.dot(X.T, dr_)      # (3.1) dL/dUr
        layer.g['dVr'] += np.dot(hp.T, dr_)     # (3.2) dL/dVr
        layer.g['dbr'] += np.sum(dr_, axis=0)   # (3.3) dL/dbr

    return None


def gru_update_parameters(layer):
    """Update parameters from gradients for layer.
    """
    for gradient in layer.g.keys():
        parameter = gradient[1:]
        # Update is driven by learning rate and gradients
        layer.p[parameter] -= layer.lrate[layer.e] * layer.g[gradient]

    return None



# EpyNN/epynn/dense/backward.py
# Related third party imports
import numpy as np

# Local application/library specific imports
from epynn.commons.maths import hadamard


def initialize_backward(layer, dX):
    """Backward cache initialization.

    :param layer: An instance of dense layer.
    :type layer: :class:`epynn.dense.models.Dense`

    :param dX: Output of backward propagation from next layer.
    :type dX: :class:`numpy.ndarray`

    :return: Input of backward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    dA = layer.bc['dA'] = dX

    return dA


def dense_backward(layer, dX):
    """Backward propagate error gradients to previous layer.
    """
    # (1) Initialize cache
    dA = initialize_backward(layer, dX)

    # (2) Gradient of the loss with respect to Z
    dZ = layer.bc['dZ'] = hadamard(
         dA,
         layer.activate(layer.fc['Z'], deriv=True)
    )    # dL/dZ

    # (3) Gradient of the loss with respect to X
    dX = layer.bc['dX'] = np.dot(dZ, layer.p['W'].T)   # dL/dX

    return dX    # To previous layer



# EpyNN/epynn/pooling/parameters.py
# Standard library imports
import math


def pooling_compute_shapes(layer, A):
    """Compute forward shapes and dimensions from input for layer.
    """
    X = A    # Input of current layer

    layer.fs['X'] = X.shape    # (m, h, w, d)

    layer.d['m'] = layer.fs['X'][0]    # Number of samples      (m)
    layer.d['h'] = layer.fs['X'][1]    # Height of features map (h)
    layer.d['w'] = layer.fs['X'][2]    # Width of features map  (w)
    layer.d['d'] = layer.fs['X'][3]    # Depth of features map  (d)

    # Output height (oh) and width (ow)
    layer.d['oh'] = math.floor((layer.d['h']-layer.d['ph']) / layer.d['sh']) + 1
    layer.d['ow'] = math.floor((layer.d['w']-layer.d['pw']) / layer.d['sw']) + 1

    return None


def pooling_initialize_parameters(layer):
    """Initialize parameters from shapes for layer.
    """
    # No parameters to initialize for Pooling layer

    return None


def pooling_compute_gradients(layer):
    """Compute gradients with respect to weight and bias for layer.
    """
    # No gradients to compute for Pooling layer

    return None


def pooling_update_parameters(layer):
    """Update parameters from gradients for layer.
    """
    # No parameters to update for Pooling layer

    return None



# EpyNN/epynn/network/forward.py


def model_forward(model, X):
    """Forward propagate input data from input to output layer.
    """
    # By convention
    A = X

    # Iterate over layers
    for layer in model.layers:

        # For learning rate schedule
        layer.e = model.e

        # Layer returns A - layer.fs, layer.fc
        A = layer.forward(A)

    return A    # To derivative of loss function



# EpyNN/epynn/flatten/models.py
# Local application/library specific imports
from epynn.commons.models import Layer
from epynn.flatten.forward import flatten_forward
from epynn.flatten.backward import flatten_backward
from epynn.flatten.parameters import (
    flatten_compute_shapes,
    flatten_initialize_parameters,
    flatten_compute_gradients,
    flatten_update_parameters
)


class Flatten(Layer):
    """
    Definition of a flatten layer prototype.
    """

    def __init__(self):
        """Initialize instance variable attributes.
        """
        super().__init__()

        self.trainable = False

        return None

    def compute_shapes(self, A):
        """Wrapper for :func:`epynn.flatten.parameters.flatten_compute_shapes()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`
        """
        flatten_compute_shapes(self, A)

        return None

    def initialize_parameters(self):
        """Wrapper for :func:`epynn.flatten.parameters.flatten_initialize_parameters()`.
        """
        flatten_initialize_parameters(self)

        return None

    def forward(self, A):
        """Wrapper for :func:`epynn.flatten.forward.flatten_forward()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`

        :return: Output of forward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        self.compute_shapes(A)
        A = flatten_forward(self, A)
        self.update_shapes(self.fc, self.fs)

        return A

    def backward(self, dX):
        """Wrapper for :func:`epynn.flatten.backward.flatten_backward()`.

        :param dX: Output of backward propagation from next layer.
        :type dX: :class:`numpy.ndarray`

        :return: Output of backward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        dX = flatten_backward(self, dX)
        self.update_shapes(self.bc, self.bs)

        return dX

    def compute_gradients(self):
        """Wrapper for :func:`epynn.flatten.parameters.flatten_compute_gradients()`. Dummy method, there are no gradients to compute in layer.
        """
        flatten_compute_gradients(self)

        return None

    def update_parameters(self):
        """Wrapper for :func:`epynn.flatten.parameters.flatten_update_parameters()`. Dummy method, there are no parameters to update in layer.
        """
        if self.trainable:
            flatten_update_parameters(self)

        return None



# EpyNN/epynn/convolution/parameters.py
# Standard library imports
import math

# Related third party imports
import numpy as np


def convolution_compute_shapes(layer, A):
    """Compute forward shapes and dimensions for layer.
    """
    X = A    # Input of current layer

    layer.fs['X'] = X.shape    # (m, h, w, d)

    layer.d['m'] = layer.fs['X'][0]     # Number of samples  (m)
    layer.d['h'] = layer.fs['X'][1]     # Height of features (h)
    layer.d['w'] = layer.fs['X'][2]     # Width of features  (w)
    layer.d['d'] = layer.fs['X'][3]     # Depth of features  (d)

    # Output height (oh) and width (ow)
    layer.d['oh'] = math.floor((layer.d['h']-layer.d['fh']) / layer.d['sh']) + 1
    layer.d['ow'] = math.floor((layer.d['w']-layer.d['fw']) / layer.d['sw']) + 1

    # Shapes for trainable parameters
    # filter_height (fh), filter_width (fw), features_depth (d), unit_filters (u)
    layer.fs['W'] = (layer.d['fh'], layer.d['fw'], layer.d['d'], layer.d['u'])
    layer.fs['b'] = (layer.d['u'], )

    return None


def convolution_initialize_parameters(layer):
    """Initialize parameters for layer.
    """
    # For linear activation of inputs (Z)
    layer.p['W'] = layer.initialization(layer.fs['W'], rng=layer.np_rng)
    layer.p['b'] = np.zeros(layer.fs['b']) # Z = X * W + b

    return None


def convolution_compute_gradients(layer):
    """Compute gradients with respect to weight and bias for layer.
    """
    # Gradients initialization with respect to parameters
    for parameter in layer.p.keys():
        gradient = 'd' + parameter
        layer.g[gradient] = np.zeros_like(layer.p[parameter])

    Xb = layer.fc['Xb']     # Input blocks of forward propagation
    dZ = layer.bc['dZ']     # Gradient of the loss with respect to Z

    # Expand dZ dimensions with respect to Xb
    dZb = dZ
    dZb = np.expand_dims(dZb, axis=3)    # (m, oh, ow, 1, u)
    dZb = np.expand_dims(dZb, axis=3)    # (m, oh, ow, 1, 1, u)
    dZb = np.expand_dims(dZb, axis=3)    # (m, oh, ow, 1, 1, 1, u)

    # (1) Gradient of the loss with respect to W, b
    dW = layer.g['dW'] = np.sum(dZb * Xb, axis=(2, 1, 0))   # (1.1) dL/dW
    db = layer.g['db'] = np.sum(dZb, axis=(2, 1, 0))        # (1.2) dL/db

    layer.g['db'] = db.squeeze() if layer.use_bias else 0.

    return None


def convolution_update_parameters(layer):
    """Update parameters for layer.
    """
    for gradient in layer.g.keys():
        parameter = gradient[1:]
        # Update is driven by learning rate and gradients
        layer.p[parameter] -= layer.lrate[layer.e] * layer.g[gradient]

    return None



# EpyNN/epynn/lstm/models.py
# Local application/library specific imports
from epynn.commons.models import Layer
from epynn.commons.maths import (
    tanh,
    sigmoid,
    orthogonal,
    clip_gradient,
    activation_tune,
)
from epynn.lstm.forward import lstm_forward
from epynn.lstm.backward import lstm_backward
from epynn.lstm.parameters import (
    lstm_compute_shapes,
    lstm_initialize_parameters,
    lstm_compute_gradients,
    lstm_update_parameters
)


class LSTM(Layer):
    """
    Definition of a LSTM layer prototype.

    :param units: Number of unit cells in LSTM layer, defaults to 1.
    :type units: int, optional

    :param activate: Non-linear activation of hidden and memory states, defaults to `tanh`.
    :type activate: function, optional

    :param activate_output: Non-linear activation of output gate, defaults to `sigmoid`.
    :type activate_output: function, optional

    :param activate_candidate: Non-linear activation of candidate, defaults to `tanh`.
    :type activate_candidate: function, optional

    :param activate_input: Non-linear activation of input gate, defaults to `sigmoid`.
    :type activate_input: function, optional

    :param activate_forget: Non-linear activation of forget gate, defaults to `sigmoid`.
    :type activate_forget: function, optional

    :param initialization: Weight initialization function for LSTM layer, defaults to `orthogonal`.
    :type initialization: function, optional

    :param clip_gradients: May prevent exploding/vanishing gradients, defaults to `False`.
    :type clip_gradients: bool, optional

    :param sequences: Whether to return only the last hidden state or the full sequence, defaults to `False`.
    :type sequences: bool, optional

    :param se_hPars: Layer hyper-parameters, defaults to `None` and inherits from model.
    :type se_hPars: dict[str, str or float] or NoneType, optional
    """

    def __init__(self,
                 unit_cells=1,
                 activate=tanh,
                 activate_output=sigmoid,
                 activate_candidate=tanh,
                 activate_input=sigmoid,
                 activate_forget=sigmoid,
                 initialization=orthogonal,
                 clip_gradients=False,
                 sequences=False,
                 se_hPars=None):
        """Initialize instance variable attributes.
        """
        super().__init__(se_hPars)

        self.d['u'] = unit_cells
        self.activate = activate
        self.activate_output = activate_output
        self.activate_candidate = activate_candidate
        self.activate_input = activate_input
        self.activate_forget = activate_forget
        self.initialization = initialization
        self.clip_gradients = clip_gradients
        self.sequences = sequences

        self.activation = {
            'activate': self.activate.__name__,
            'activate_output': self.activate_output.__name__,
            'activate_candidate': self.activate_candidate.__name__,
            'activate_input': self.activate_input.__name__,
            'activate_forget': self.activate_forget.__name__,
        }
        self.trainable = True

        return None

    def compute_shapes(self, A):
        """Is a wrapper for :func:`epynn.lstm.parameters.lstm_compute_shapes()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`
        """
        lstm_compute_shapes(self, A)

        return None

    def initialize_parameters(self):
        """Is a wrapper for :func:`epynn.lstm.parameters.lstm_initialize_parameters()`.
        """
        lstm_initialize_parameters(self)

        return None

    def forward(self, A):
        """Is a wrapper for :func:`epynn.lstm.forward.lstm_forward()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`

        :return: Output of forward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        self.compute_shapes(A)
        activation_tune(self.se_hPars)
        A = self.fc['A'] = lstm_forward(self, A)
        self.update_shapes(self.fc, self.fs)

        return A

    def backward(self, dX):
        """Is a wrapper for :func:`epynn.lstm.backward.lstm_backward()`.

        :param dX: Output of backward propagation from next layer.
        :type dX: :class:`numpy.ndarray`

        :return: Output of backward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        activation_tune(self.se_hPars)
        dX = lstm_backward(self, dX)
        self.update_shapes(self.bc, self.bs)

        return dX

    def compute_gradients(self):
        """Is a wrapper for :func:`epynn.lstm.parameters.lstm_compute_gradients()`.
        """
        lstm_compute_gradients(self)

        if self.clip_gradients:
            clip_gradient(self)

        return None

    def update_parameters(self):
        """Is a wrapper for :func:`epynn.lstm.parameters.lstm_update_parameters()`.
        """
        if self.trainable:
            lstm_update_parameters(self)

        return None



# EpyNN/epynn/gru/backward.py
# Related third party imports
import numpy as np


def initialize_backward(layer, dX):
    """Backward cache initialization.

    :param layer: An instance of GRU layer.
    :type layer: :class:`epynn.gru.models.GRU`

    :param dX: Output of backward propagation from next layer.
    :type dX: :class:`numpy.ndarray`

    :return: Input of backward propagation for current layer.
    :rtype: :class:`numpy.ndarray`

    :return: Next hidden state initialized with zeros.
    :rtype: :class:`numpy.ndarray`
    """
    if layer.sequences:
        dA = dX                         # Full length sequence
    elif not layer.sequences:
        dA = np.zeros(layer.fs['h'])    # Empty full length sequence
        dA[:, -1] = dX                  # Assign to last index

    cache_keys = ['dh_', 'dh', 'dhn', 'dhh_', 'dz_', 'dr_']
    layer.bc.update({k: np.zeros(layer.fs['h']) for k in cache_keys})

    layer.bc['dA'] = dA
    layer.bc['dX'] = np.zeros(layer.fs['X'])    # To previous layer

    dh = layer.bc['dh'][:, 0]                   # To previous step

    return dA, dh


def gru_backward(layer, dX):
    """Backward propagate error gradients to previous layer.
    """
    # (1) Initialize cache and hidden state gradients
    dA, dh = initialize_backward(layer, dX)

    # Reverse iteration over sequence steps
    for s in reversed(range(layer.d['s'])):

        # (2s) Slice sequence (m, s, u) w.r.t step
        dA = layer.bc['dA'][:, s]          # dL/dA

        # (3s) Retrieve previous hidden state
        dhn = layer.bc['dhn'][:, s] = dh   # dL/dhn

        # (4s) Gradient of the loss w.r.t hidden state h_
        dh_ = layer.bc['dh_'][:, s] = (
            (dA + dhn)
        )   # dL/dh_

        # (5s) Gradient of the loss w.r.t hidden hat hh_
        dhh_ = layer.bc['dhh_'][:, s] = (
            dh_
            * (1-layer.fc['z'][:, s])
            * layer.activate(layer.fc['hh_'][:, s], deriv=True)
        )   # dL/dhh_

        # (6s) Gradient of the loss w.r.t update gate z_
        dz_ = layer.bc['dz_'][:, s] = (
            dh_
            * (layer.fc['hp'][:, s]-layer.fc['hh'][:, s])
            * layer.activate_update(layer.fc['z_'][:, s], deriv=True)
        )   # dL/dz_

        # (7s) Gradient of the loss w.r.t reset gate
        dr_ = layer.bc['dr_'][:, s] = (
            np.dot(dhh_, layer.p['Vhh'].T)
            * layer.fc['hp'][:, s]
            * layer.activate_reset(layer.fc['r_'][:, s], deriv=True)
        )   # dL/dr_

        # (8s) Gradient of the loss w.r.t previous hidden state
        dh = layer.bc['dh'][:, s] = (
            np.dot(dhh_, layer.p['Vhh'].T) * layer.fc['r'][:, s]
            + np.dot(dz_, layer.p['Vz'].T) + dh_ * layer.fc['z'][:, s]
            + np.dot(dr_, layer.p['Vr'].T)
        )   # dL/dh

        # (9s) Gradient of the loss w.r.t to X
        dX = layer.bc['dX'][:, s] = (
            np.dot(dhh_, layer.p['Uhh'].T)
            + np.dot(dz_, layer.p['Uz'].T)
            + np.dot(dr_, layer.p['Ur'].T)
        )   # dL/dX

    dX = layer.bc['dX']

    return dX    # To previous layer



# EpyNN/epynn/rnn/backward.py
# Related third party imports
import numpy as np


def initialize_backward(layer, dX):
    """Backward cache initialization.

    :param layer: An instance of RNN layer.
    :type layer: :class:`epynn.rnn.models.RNN`

    :param dX: Output of backward propagation from next layer.
    :type dX: :class:`numpy.ndarray`

    :return: Input of backward propagation for current layer.
    :rtype: :class:`numpy.ndarray`

    :return: Next hidden state initialized with zeros.
    :rtype: :class:`numpy.ndarray`
    """
    if layer.sequences:
        dA = dX                         # Full length sequence
    elif not layer.sequences:
        dA = np.zeros(layer.fs['h'])    # Empty full length sequence
        dA[:, -1] = dX                  # Assign to last index

    cache_keys = ['dh_', 'dh', 'dhn']
    layer.bc.update({k: np.zeros(layer.fs['h']) for k in cache_keys})

    layer.bc['dA'] = dA
    layer.bc['dX'] = np.zeros(layer.fs['X'])    # To previous layer

    dh = layer.bc['dh'][:, 0]                   # To previous step

    return dA, dh


def rnn_backward(layer, dX):
    """Backward propagate error gradients to previous layer.
    """
    # (1) Initialize cache and hidden state gradient
    dA, dh = initialize_backward(layer, dX)

    # Reverse iteration over sequence steps
    for s in reversed(range(layer.d['s'])):

        # (2s) Slice sequence (m, s, u) w.r.t step
        dA = layer.bc['dA'][:, s]          # dL/dA

        # (3s) Gradient of the loss w.r.t. next hidden state
        dhn = layer.bc['dhn'][:, s] = dh   # dL/dhn

        # (4s) Gradient of the loss w.r.t hidden state h_
        dh_ = layer.bc['dh_'][:, s] = (
            (dA + dhn)
            * layer.activate(layer.fc['h_'][:, s], deriv=True)
        )   # dL/dh_ - To parameters gradients

        # (5s) Gradient of the loss w.r.t hidden state h
        dh = layer.bc['dh'][:, s] = (
            np.dot(dh_, layer.p['V'].T)
        )   # dL/dh - To previous step

        # (6s) Gradient of the loss w.r.t X
        dX = layer.bc['dX'][:, s] = (
            np.dot(dh_, layer.p['U'].T)
        )   # dL/dX - To previous layer

    dX = layer.bc['dX']

    return dX    # To previous layer



# EpyNN/epynn/lstm/backward.py
# Related third party imports
import numpy as np


def initialize_backward(layer, dX):
    """Backward cache initialization.

    :param layer: An instance of LSTM layer.
    :type layer: :class:`epynn.lstm.models.LSTM`

    :param dX: Output of backward propagation from next layer.
    :type dX: :class:`numpy.ndarray`

    :return: Input of backward propagation for current layer.
    :rtype: :class:`numpy.ndarray`

    :return: Next hidden state initialized with zeros.
    :rtype: :class:`numpy.ndarray`

    :return: Next memory state initialized with zeros.
    :rtype: :class:`numpy.ndarray`
    """
    if layer.sequences:
        dA = dX                         # Full length sequence
    elif not layer.sequences:
        dA = np.zeros(layer.fs['h'])    # Empty full length sequence
        dA[:, -1] = dX                  # Assign to last index

    cache_keys = [
        'dh_', 'dh', 'dhn',
        'dC_', 'dC', 'dCn',
        'do_', 'dg_', 'di_', 'df_', 'dz_'
    ]

    layer.bc.update({k: np.zeros(layer.fs['h']) for k in cache_keys})

    layer.bc['dA'] = dA
    layer.bc['dX'] = np.zeros(layer.fs['X'])    # To previous layer

    dh = layer.bc['dh'][:, 0]                   # To previous step
    dC = layer.bc['dC'][:, 0]                   # To previous step

    return dA, dh, dC


def lstm_backward(layer, dX):
    """Backward propagate error gradients to previous layer.
    """
    # (1) Initialize cache, hidden and memory state gradients
    dA, dh, dC = initialize_backward(layer, dX)

    # Reverse iteration over sequence steps
    for s in reversed(range(layer.d['s'])):

        # (2s) Slice sequence (m, s, u) w.r.t step
        dA = layer.bc['dA'][:, s]          # dL/dA

        # (3s) Gradient of the loss w.r.t. next states
        dhn = layer.bc['dhn'][:, s] = dh   # (3.1) dL/dhn
        dCn = layer.bc['dCn'][:, s] = dC   # (3.2) dL/dCn

        # (4s) Gradient of the loss w.r.t hidden state h_
        dh_ = layer.bc['dh_'][:, s] = (
            (dA + dhn)
        )   # dL/dh_

        # (5s) Gradient of the loss w.r.t memory state C_
        dC_ = layer.bc['dC_'][:, s] = (
            dh_
            * layer.fc['o'][:, s]
            * layer.activate(layer.fc['C_'][:, s], deriv=True)
            + dCn
        )   # dL/dC_

        # (6s) Gradient of the loss w.r.t output gate o_
        do_ = layer.bc['do_'][:, s] = (
            dh_
            * layer.fc['C'][:, s]
            * layer.activate_output(layer.fc['o_'][:, s], deriv=True)
        )   # dL/do_

        # (7s) Gradient of the loss w.r.t candidate g_
        dg_ = layer.bc['dg_'][:, s] = (
            dC_
            * layer.fc['i'][:, s]
            * layer.activate_candidate(layer.fc['g_'][:, s], deriv=True)
        )   # dL/dg_

        # (8s) Gradient of the loss w.r.t input gate i_
        di_ = layer.bc['di_'][:, s] = (
            dC_
            * layer.fc['g'][:, s]
            * layer.activate_input(layer.fc['i_'][:, s], deriv=True)
        )   # dL/di_

        # (9s) Gradient of the loss w.r.t forget gate f_
        df_ = layer.bc['df_'][:, s] = (
            dC_
            * layer.fc['Cp_'][:, s]
            * layer.activate_forget(layer.fc['f_'][:, s], deriv=True)
        )   # dL/df_

        # (10s) Gradient of the loss w.r.t memory state C
        dC = layer.bc['dC'][:, s] = (
            dC_
            * layer.fc['f'][:, s]
        )   # dL/dC

        # (11s) Gradient of the loss w.r.t hidden state h
        dh = layer.bc['dh'][:, s] = (
            np.dot(do_, layer.p['Vo'].T)
            + np.dot(dg_, layer.p['Vg'].T)
            + np.dot(di_, layer.p['Vi'].T)
            + np.dot(df_, layer.p['Vf'].T)
        )   # dL/dh

        # (12s) Gradient of the loss w.r.t hidden state X
        dX = layer.bc['dX'][:, s] = (
            np.dot(dg_, layer.p['Ug'].T)
            + np.dot(do_, layer.p['Uo'].T)
            + np.dot(di_, layer.p['Ui'].T)
            + np.dot(df_, layer.p['Uf'].T)
        )   # dL/dX

    dX = layer.bc['dX']

    return dX    # To previous layer



# EpyNN/epynn/pool/models.py
# Related third party imports
import numpy as np

# Local application/library specific imports
from epynn.commons.models import Layer
from epynn.pooling.forward import pooling_forward
from epynn.pooling.backward import pooling_backward
from epynn.pooling.parameters import (
    pooling_compute_shapes,
    pooling_initialize_parameters,
    pooling_compute_gradients,
    pooling_update_parameters
)


class Pooling(Layer):
    """
    Definition of a pooling layer prototype.

    :param pool_size: Height and width for pooling window, defaults to `(2, 2)`.
    :type pool_size: int or tuple[int], optional

    :param strides: Height and width to shift the pooling window by, defaults to `None` which equals `pool_size`.
    :type strides: int or tuple[int], optional

    :param pool: Pooling activation of units, defaults to :func:`np.max`. Use one of min or max pooling.
    :type pool: function, optional
    """

    def __init__(self,
                 pool_size=(2, 2),
                 strides=None,
                 pool=np.max):

        super().__init__()

        pool_size = pool_size if isinstance(pool_size, tuple) else (pool_size, pool_size)
        strides = strides if isinstance(strides, tuple) else pool_size

        self.d['ph'], self.d['pw'] = pool_size
        self.d['sh'], self.d['sw'] = strides
        self.pool = pool

        self.activation = { 'pool': self.pool.__name__ }
        self.trainable = False

        return None

    def compute_shapes(self, A):
        """Wrapper for :func:`epynn.pooling.parameters.pooling_compute_shapes()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`
        """
        pooling_compute_shapes(self, A)

        return None

    def initialize_parameters(self):
        """Wrapper for :func:`epynn.pooling.parameters.initialize_parameters()`.
        """
        pooling_initialize_parameters(self)

        return None

    def forward(self, A):
        """Wrapper for :func:`epynn.pooling.forward.pooling_forward()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`

        :return: Output of forward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        self.compute_shapes(A)
        A = pooling_forward(self, A)
        self.update_shapes(self.fc, self.fs)

        return A

    def backward(self, dX):
        """Wrapper for :func:`epynn.pooling.backward.pooling_backward()`.

        :param dX: Output of backward propagation from next layer.
        :type dX: :class:`numpy.ndarray`

        :return: Output of backward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        dX = pooling_backward(self, dX)
        self.update_shapes(self.bc, self.bs)

        return dX

    def compute_gradients(self):
        """Wrapper for :func:`epynn.pooling.parameters.pooling_compute_gradients()`. Dummy method, there are no gradients to compute in layer.
        """
        pooling_compute_gradients(self)

        return None

    def update_parameters(self):
        """Wrapper for :func:`epynn.pooling.parameters.pooling_update_parameters()`. Dummy method, there are no parameters to update in layer.
        """
        if self.trainable:
            pooling_update_parameters(self)

        return None



# EpyNN/epynn/commons/metrics.py
# Related third party imports
import numpy as np


def metrics_functions(key=None):
    """Callback function for metrics.

    :param key: Name of the metrics function, defaults to `None` which returns all functions.
    :type key: str, optional

    :return: Metrics functions or computed metrics.
    :rtype: dict[str: function] or :class:`numpy.ndarray`
    """
    metrics = {
        'accuracy': accuracy,
        'recall': recall,
        'precision': precision,
        'fscore': fscore,
        'specificity': specificity,
        'NPV': NPV,
    }
    # If key provided, returns output of function
    if key:
        metrics = metrics[key]

    return metrics


def accuracy(Y, A):
    """Accuracy of prediction.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :return: Accuracy for each sample.
    :rtype: :class:`numpy.ndarray`
    """
    encoded = (Y.shape[1] > 1)

    P = np.argmax(A, axis=1) if encoded else np.around(A)
    y = np.argmax(Y, axis=1) if encoded else Y

    accuracy = (P == y)

    return accuracy


def recall(Y, A):
    """Fraction of positive instances retrieved over the total.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :return: Recall.
    :rtype: :class:`numpy.ndarray`
    """
    encoded = (Y.shape[1] > 1)    # Check if one-hot encoding of labels

    P = np.argmax(A, axis=1) if encoded else np.around(A)
    y = np.argmax(Y, axis=1) if encoded else Y

    tp = np.sum(np.where((P==0) & (y==0), 1, 0))    # True positive
    fp = np.sum(np.where((P==0) & (y==1), 1, 0))    # False positive
    tn = np.sum(np.where((P==1) & (y==1), 1, 0))    # True negative
    fn = np.sum(np.where((P==1) & (y==0), 1, 0))    # False negative

    recall = (tp / (tp+fn))

    return recall


def precision(Y, A):
    """Fraction of positive samples among retrieved instances.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :return: Precision.
    :rtype: :class:`numpy.ndarray`
    """
    encoded = (Y.shape[1] > 1)    # Check if one-hot encoding of labels

    P = np.argmax(A, axis=1) if encoded else np.around(A)
    y = np.argmax(Y, axis=1) if encoded else Y

    tp = np.sum(np.where((P==0) & (y==0), 1, 0))    # True positive
    fp = np.sum(np.where((P==0) & (y==1), 1, 0))    # False positive
    tn = np.sum(np.where((P==1) & (y==1), 1, 0))    # True negative
    fn = np.sum(np.where((P==1) & (y==0), 1, 0))    # False negative

    precision = (tp / (tp+fp))

    return precision


def NPV(Y, A):
    """Fraction of negative samples among excluded instances.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :return: Negative Predictive Value.
    :rtype: :class:`numpy.ndarray`
    """
    encoded = (Y.shape[1] > 1)    # Check if one-hot encoding of labels

    P = np.argmax(A, axis=1) if encoded else np.around(A)
    y = np.argmax(Y, axis=1) if encoded else Y

    tp = np.sum(np.where((P==0) & (y==0), 1, 0))    # True positive
    fp = np.sum(np.where((P==0) & (y==1), 1, 0))    # False positive
    tn = np.sum(np.where((P==1) & (y==1), 1, 0))    # True negative
    fn = np.sum(np.where((P==1) & (y==0), 1, 0))    # False negative

    npv = (tn / (tn+fn))

    return npv


def fscore(Y, A):
    """F-Score that is the harmonic mean of recall and precision.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :return: F-score.
    :rtype: :class:`numpy.ndarray`
    """
    encoded = (Y.shape[1] > 1)    # Check if one-hot encoding of labels

    P = np.argmax(A, axis=1) if encoded else np.around(A)
    y = np.argmax(Y, axis=1) if encoded else Y

    tp = np.sum(np.where((P==0) & (y==0), 1, 0))    # True positive
    fp = np.sum(np.where((P==0) & (y==1), 1, 0))    # False positive
    tn = np.sum(np.where((P==1) & (y==1), 1, 0))    # True negative
    fn = np.sum(np.where((P==1) & (y==0), 1, 0))    # False negative

    fscore = (tp / (tp + 0.5*(fp+fn)))

    return fscore

 
def specificity(Y, A):
    """Fraction of negative samples among excluded instances.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :return: Specificity.
    :rtype: :class:`numpy.ndarray`
    """
    encoded = (Y.shape[1] > 1)    # Check if one-hot encoding of labels

    P = np.argmax(A, axis=1) if encoded else np.around(A)
    y = np.argmax(Y, axis=1) if encoded else Y

    tp = np.sum(np.where((P==0) & (y==0), 1, 0))    # True positive
    fp = np.sum(np.where((P==0) & (y==1), 1, 0))    # False positive
    tn = np.sum(np.where((P==1) & (y==1), 1, 0))    # True negative
    fn = np.sum(np.where((P==1) & (y==0), 1, 0))    # False negative

    specificity = (tn / (tn+fp))

    return specificity



# EpyNN/epynn/rnn/models.py
# Local application/library specific imports
from epynn.commons.models import Layer
from epynn.commons.maths import (
    tanh,
    xavier,
    clip_gradient,
    activation_tune,
)
from epynn.rnn.forward import rnn_forward
from epynn.rnn.backward import rnn_backward
from epynn.rnn.parameters import (
    rnn_compute_shapes,
    rnn_initialize_parameters,
    rnn_compute_gradients,
    rnn_update_parameters
)


class RNN(Layer):
    """
    Definition of a RNN layer prototype.

    :param units: Number of unit cells in RNN layer, defaults to 1.
    :type units: int, optional

    :param activate: Non-linear activation of hidden state, defaults to `tanh`.
    :type activate: function, optional

    :param initialization: Weight initialization function for RNN layer, defaults to `xavier`.
    :type initialization: function, optional

    :param clip_gradients: May prevent exploding/vanishing gradients, defaults to `False`.
    :type clip_gradients: bool, optional

    :param sequences: Whether to return only the last hidden state or the full sequence, defaults to `False`.
    :type sequences: bool, optional

    :param se_hPars: Layer hyper-parameters, defaults to `None` and inherits from model.
    :type se_hPars: dict[str, str or float] or NoneType, optional
    """

    def __init__(self,
                 unit_cells=1,
                 activate=tanh,
                 initialization=xavier,
                 clip_gradients=True,
                 sequences=False,
                 se_hPars=None):
        """Initialize instance variable attributes.
        """
        super().__init__(se_hPars)

        self.d['u'] = unit_cells
        self.activate = activate
        self.initialization = initialization
        self.clip_gradients = clip_gradients
        self.sequences = sequences

        self.activation = { 'activate': self.activate.__name__ }
        self.trainable = True

        return None

    def compute_shapes(self, A):
        """Wrapper for :func:`epynn.rnn.parameters.rnn_compute_shapes()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`
        """
        rnn_compute_shapes(self, A)

        return None

    def initialize_parameters(self):
        """Wrapper for :func:`epynn.rnn.parameters.rnn_initialize_parameters()`.
        """
        rnn_initialize_parameters(self)

        return None

    def forward(self, A):
        """Wrapper for :func:`epynn.rnn.forward.rnn_forward()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`

        :return: Output of forward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        self.compute_shapes(A)
        activation_tune(self.se_hPars)
        A = self.fc['A'] = rnn_forward(self, A)
        self.update_shapes(self.fc, self.fs)

        return A

    def backward(self, dX):
        """Wrapper for :func:`epynn.rnn.backward.rnn_backward()`.

        :param dX: Output of backward propagation from next layer.
        :type dX: :class:`numpy.ndarray`

        :return: Output of backward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        activation_tune(self.se_hPars)
        dX = rnn_backward(self, dX)
        self.update_shapes(self.bc, self.bs)

        return dX

    def compute_gradients(self):
        """Wrapper for :func:`epynn.rnn.parameters.rnn_compute_gradients()`.
        """
        rnn_compute_gradients(self)

        if self.clip_gradients:
            clip_gradient(self)

        return None

    def update_parameters(self):
        """Wrapper for :func:`epynn.rnn.parameters.rnn_update_parameters()`.
        """
        if self.trainable:
            rnn_update_parameters(self)

        return None



# EpyNN/epynn/commons/maths.py
# Related third party imports
import numpy as np


# To prevent from divide floatting points errors
E_SAFE = 1e-16

 
def activation_tune(se_hPars):
    """Set layer's hyperparameters as temporary globals.

    For each forward and backward pass the function is called from within the layer.

    :param se_hPars: Local hyperparameters for layers.
    :type se_hPars: dict[str, str or float]
    """
    # Declare global
    global layer_hPars
    # Set global
    layer_hPars = se_hPars

    return None


def hadamard(dA, dLinear):
    """Element-wise matrix multiplication with support for softmax derivative.

    This is implemented for Dense layer and is compatible with other layers satisfying requirements.

    :param dA: Input of backward propagation of shape (m, n).
    :type dA: :class:`numpy.ndarray`

    :param dLinear: Linear activation product passed through the derivative of the non-linear activation function with shape (m, n) or (m, n, n).
    :type dLinear: :class:`numpy.ndarray`
    """

    # Non-softmax processing
    if dLinear.ndim == 2:
        dZ = dA * dLinear

    # Softmax processing
    elif dLinear.ndim == 3:
        dA = np.expand_dims(dA, 2)
        dZ = dA * dLinear
        dZ = np.sum(dZ, axis=1)

    return dZ


### Activation functions and derivatives

# Identity function

def identity(x, deriv=False):
    """Compute identity activation or derivative.

    Note this is for testing purpose, cannot be used with backpropagation.

    :param x: Input array to pass in function.
    :type x: class:`numpy.ndarray`

    :param deriv: To compute derivative, defaults to False.
    :type deriv: bool, optional

    :return: Output array passed in function.
    :rtype: :class:`numpy.ndarray`
    """
    if not deriv:
        pass

    elif deriv:
        x = np.ones_like(x)

    return x


# Rectifier Linear Unit (ReLU)

def relu(x, deriv=False):
    """Compute ReLU activation or derivative.

    :param x: Input array to pass in function.
    :type x: class:`numpy.ndarray`

    :param deriv: To compute derivative, defaults to False.
    :type deriv: bool, optional

    :return: Output array passed in function.
    :rtype: :class:`numpy.ndarray`
    """
    if not deriv:
        x = np.maximum(0, x)

    elif deriv:
        x = np.greater(x, 0).astype(int)

    return x


# Leaky Rectifier Linear Unit (LReLU)

def lrelu(x, deriv=False):
    """Compute LReLU activation or derivative.

    :param x: Input array to pass in function.
    :type x: class:`numpy.ndarray`

    :param deriv: To compute derivative, defaults to False.
    :type deriv: bool, optional

    :return: Output array passed in function.
    :rtype: :class:`numpy.ndarray`
    """
    # Retrieve alpha from layers hyperparameters (temporary globals)
    a = layer_hPars['LRELU_alpha']

    if not deriv:
        x = np.maximum(a * x, x)

    elif deriv:
        x = np.where(x > 0, 1, a)

    return x


# Exponential Linear Unit (ELU)

def elu(x, deriv=False):
    """Compute ELU activation or derivative.

    :param x: Input array to pass in function.
    :type x: class:`numpy.ndarray`

    :param deriv: To compute derivative, defaults to False.
    :type deriv: bool, optional

    :return: Output array passed in function.
    :rtype: :class:`numpy.ndarray`
    """
    # Retrieve alpha from layers hyperparameters (temporary globals)
    a = layer_hPars['ELU_alpha']

    if not deriv:
        x = np.where(x > 0, x, a * (np.exp(x, where=x<=0)-1))

    elif deriv:
        x = np.where(x > 0, 1, elu(x) + a)

    return x


# Sigmoid (σ)

def sigmoid(x, deriv=False):
    """Compute Sigmoid activation or derivative.

    :param x: Input array to pass in function.
    :type x: class:`numpy.ndarray`

    :param deriv: To compute derivative, defaults to False.
    :type deriv: bool, optional

    :return: Output array passed in function.
    :rtype: :class:`numpy.ndarray`
    """
    if not deriv:
        # Numerically stable version of sigmoid function
        x = np.where(
            x >= 0, # condition
            1 / (1+np.exp(-x)), # For positive values
            np.exp(x) / (1+np.exp(x)) # For negative values
        )

    elif deriv:
        x = sigmoid(x) * (1-sigmoid(x))

    return x


# Hyperbolic tangent (tanh)

def tanh(x, deriv=False):
    """Compute tanh activation or derivative.

    :param x: Input array to pass in function.
    :type x: class:`numpy.ndarray`

    :param deriv: To compute derivative, defaults to False.
    :type deriv: bool, optional

    :return: Output array passed in function.
    :rtype: :class:`numpy.ndarray`
    """
    if not deriv:
        x = (np.exp(x)-np.exp(-x)) / (np.exp(x)+np.exp(-x))

    elif deriv:
        x = 1 - tanh(x)**2

    return x


# Softmax

def softmax(x, deriv=False):
    """Compute softmax activation or derivative.

    For Dense layer only.

    For other layers, you can change element-wise matrix multiplication
    operator '*' by :func:`epynn.maths.hadamard` which handles
    the softmax derivative jacobian matrix.

    :param x: Input array to pass in function.
    :type x: class:`numpy.ndarray`

    :param deriv: To compute derivative, defaults to False.
    :type deriv: bool, optional

    :return: Output array passed in function.
    :rtype: :class:`numpy.ndarray`
    """
    # Retrieve temperature from layers hyperparameters (temporary globals)
    T = layer_hPars['softmax_temperature']

    if not deriv:
        # Numerically stable version of softmax function
        x_safe = x - np.max(x, axis=1, keepdims=True)

        x_exp = np.exp(x_safe / T)
        x_sum = np.sum(x_exp, axis=1, keepdims=True)

        x = x_exp / x_sum

    elif deriv:

        x = np.array([np.diag(x) - np.outer(x, x) for x in softmax(x)])

    return x


### Weight initialization

# Xavier

def xavier(shape, rng=np.random):
    """Xavier Normal Distribution initialization for weight array.

    :param shape: Shape of weight array.
    :type shape: tuple[int]

    :param rng: Pseudo-random number generator, defaults to `np.random`.
    :type rng: :class:`numpy.random`

    :return: Initialized weight array.
    :rtype: :class:`numpy.ndarray`
    """
    W = rng.standard_normal(shape)        # Normal distribution, zero-centered
    W *= np.sqrt(2 / sum(list(shape)))    # Scale

    return W


# Orthogonal

def orthogonal(shape, rng=np.random):
    """Orthogonal initialization for weight array.

    :param shape: Shape of weight array.
    :type shape: tuple[int]

    :param rng: Pseudo-random number generator, defaults to `np.random`.
    :type rng: :class:`numpy.random`

    :return: Initialized weight array.
    :rtype: :class:`numpy.ndarray`
    """
    W = rng.standard_normal(shape)

    W = W.T if shape[0] < shape[1] else W

    # Compute QR factorization
    q, r = np.linalg.qr(W)

    # Make Q uniform according
    d = np.diag(r, 0)
    ph = np.sign(d)
    q *= ph

    W = q.T if shape[0] < shape[1] else q

    return W


### Gradients clipping


def clip_gradient(layer, max_norm=0.25):
    """Clip to avoid vanishing or exploding gradients.

    :param layer: An instance of trainable layer.
    :type layer: Object

    :param max_norm: Maximal clipping coefficient allowed, defaults to 0.25.
    :type max_norm: float, optional
    """
    total_norm = 0

    # Compute L2 norm squared for each gradient
    for grad in layer.g.values():
        grad_norm = np.sum(np.power(grad + E_SAFE, 2))
        total_norm += grad_norm    # Add to total norm

    total_norm = np.sqrt(total_norm)

    # Compute clipping coeficient
    clip_coef = max_norm / (total_norm+E_SAFE)

    # Clip the gradient if norm is greater than max norm
    if clip_coef < 1:
        for g in layer.g.keys():
            layer.g[g] *= clip_coef

    return None



# EpyNN/epynn/commons/loss.py
# Related third party imports
import numpy as np


# To prevent from divide floatting points errors
E_SAFE = 1e-16
 

def loss_functions(key=None, output_activation=None):
    """Callback function for loss.

    :param key: Name of the loss function, defaults to `None` which returns all functions.
    :type key: str, optional

    :param output_activation: Name of the activation function for output layer.
    :type output_activation: str, optional

    :raises Exception: If key is `CCE` and output activation is different from softmax.

    :raises Exception: If key is either `CCE`, `BCE` or `MSLE` and output activation is tanh.

    :return: Loss functions or computed loss.
    :rtype: dict[str, function] or :class:`numpy.ndarray`
    """
    loss = {
        'MSE': MSE,
        'MAE': MAE,
        'MSLE': MSLE,
        'CCE': CCE,
        'BCE': BCE,
    }
    # If key provided, returns output of function
    if key:
        loss = loss[key]

    if key == 'CCE' and output_activation != 'softmax':
        raise Exception('CCE can not be used with %s activation, \
                        please use softmax instead.' % output_activation)

    if key in ['CCE', 'BCE', 'MSLE'] and output_activation == 'tanh':
        raise Exception('%s contains log() not be used with %s activation, \
                         please change.' % (key, output_activation))

    return loss


def MSE(Y, A, deriv=False):
    """Mean Squared Error.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :param deriv: To compute the derivative.
    :type deriv: bool, optional

    :return: Loss.
    :rtype: :class:`numpy.ndarray`
    """
    U = A.shape[1]    # Number of output nodes

    if not deriv:
        loss =  1. / U * np.sum((Y - A)**2, axis=1)

    elif deriv:
        loss = -2. / U * (Y-A)

    return loss


def MAE(Y, A, deriv=False):
    """Mean Absolute Error.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :param deriv: To compute the derivative.
    :type deriv: bool, optional

    :return: Loss.
    :rtype: :class:`numpy.ndarray`
    """
    U = A.shape[1]    # Number of output nodes

    if not deriv:
        loss =  1. / U * np.sum(np.abs(Y-A), axis=1)

    elif deriv:
        loss = -1. / U * (Y-A) / (np.abs(Y-A)+E_SAFE)

    return loss


def MSLE(Y, A, deriv=False):
    """Mean Squared Logarythmic Error.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :param deriv: To compute the derivative.
    :type deriv: bool, optional

    :return: Loss.
    :rtype: :class:`numpy.ndarray`
    """
    U = A.shape[1]    # Number of output nodes

    if not deriv:
        loss = 1. / U * np.sum(np.square(np.log1p(Y) - np.log1p(A)), axis=1)

    elif deriv:
        loss = -2. / U * (np.log1p(Y) - np.log1p(A)) / (A + 1.)

    return loss


def CCE(Y, A, deriv=False):
    """Categorical Cross-Entropy.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :param deriv: To compute the derivative.
    :type deriv: bool, optional

    :return: Loss.
    :rtype: :class:`numpy.ndarray`
    """
    U = A.shape[1]    # Number of output nodes

    if not deriv:
        loss = -1. * np.sum(Y * np.log(A+E_SAFE), axis=1)

    elif deriv:
        loss = -1. * (Y / A)

    return loss


def BCE(Y, A, deriv=False):
    """Binary Cross-Entropy.

    :param Y: True labels for a set of samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation.
    :type A: :class:`numpy.ndarray`

    :param deriv: To compute the derivative.
    :type deriv: bool, optional

    :return: Loss.
    :rtype: :class:`numpy.ndarray`
    """
    U = A.shape[1]    # Number of output nodes

    if not deriv:
        loss = -1. / U * np.sum(Y*np.log(A+E_SAFE) + (1-Y)*np.log((1-A)+E_SAFE), axis=1)

    elif deriv:
        loss = 1. / U * (A-Y) / (A - A*A + E_SAFE)

    return loss



# EpyNN/epynn/commons/schedule.py
# Related third party imports
import numpy as np
 

def schedule_functions(schedule, hPars):
    """Roots hyperparameters to relevant scheduler.

    :param schedule: Schedule mode.
    :type schedule: str

    :param hPars: Contains hyperparameters.
    :type hPars: tuple[int or float]

    :return: Scheduled learning rate.
    :rtype: list[float]
    """
    schedulers = {
        'exp_decay': exp_decay,
        'lin_decay': lin_decay,
        'steady': steady,
    }

    lrate = schedulers[schedule](hPars)

    return lrate


def exp_decay(hPars):
    """Exponential decay schedule for learning rate.

    :param hPars: Contains hyperparameters.
    :type hPars: tuple[int or float]

    :return: Scheduled learning rate.
    :rtype: list[float]
    """
    e, lr, n, k, d, epc = hPars

    lrate = [lr * (1-d) ** (x//epc) * np.exp(-(x%epc) * k) for x in range(e)]

    return lrate


def lin_decay(hPars):
    """Linear decay schedule for learning rate.

    :param hPars: Contains hyperparameters.
    :type hPars: tuple[int or float]

    :return: Scheduled learning rate.
    :rtype: list[float]
    """
    e, lr, n, k, d, epc = hPars

    lrate = [lr / (1 + k*100*x) for x in range(e)]

    return lrate


def steady(hPars):
    """Steady schedule for learning rate.

    :param hPars: Contains hyperparameters.
    :type hPars: tuple[int or float]

    :return: Scheduled learning rate.
    :rtype: list[float]
    """
    e, lr, n, k, d, epc = hPars

    lrate = [lr for x in range(e)]

    return lrate



# EpyNN/epynn/embedding/dataset.py
# Standard library imports
import warnings

# Related third party imports
import numpy as np

# Local application/library specific imports
from epynn.commons.io import (
    encode_dataset,
    scale_features,
    index_elements_auto,
)
from epynn.commons.models import dataSet


def embedding_check(X_data, Y_data=None, X_scale=False):
    """Pre-processing.

    :param X_data: Set of sample features.
    :type encode: list[list] or :class:`numpy.ndarray`

    :param Y_data: Set of samples label.
    :type encode: list[list[int] or int] or :class:`numpy.ndarray`, optional

    :param X_scale: Set to True to normalize sample features within [0, 1].
    :type X_scale: bool, optional

    :return: Sample features and label.
    :rtype: tuple[:class:`numpy.ndarray`]
    """
    if X_scale:
        # Array-wide normalization in [0, 1]
        X_data = scale_features(X_data)

    X_data = np.array(X_data)

    Y_data = np.array(Y_data)

    return X_data, Y_data


def embedding_encode(layer, X_data, Y_data, X_encode, Y_encode):
    """One-hot encoding for samples features and label.

    :param layer: An instance of the :class:`epynn.embedding.models.Embedding`
    :type layer: :class:`epynn.embedding.models.Embedding`

    :param X_data: Set of sample features.
    :type encode: list[list] or :class:`numpy.ndarray`

    :param Y_data: Set of samples label.
    :type encode: list[list[int] or int] or :class:`numpy.ndarray`

    :param X_encode: Set to True to one-hot encode features.
    :type encode: bool

    :param Y_encode: Set to True to one-hot encode labels.
    :type encode: bool

    :return: Encoded set of sample features, if applicable.
    :rtype : :class:`numpy.ndarray`

    :return: Encoded set of sample label, if applicable.
    :rtype : :class:`numpy.ndarray`
    """
    # Features one-hot encoding
    if X_encode:
        layer.e2i, layer.i2e, layer.d['e'] = index_elements_auto(X_data)
        X_data = encode_dataset(X_data, layer.e2i, layer.d['e'])
    # Label one-hot encoding
    if Y_encode:
        num_classes = len(list(set(Y_data.flatten())))
        Y_data = np.eye(num_classes)[Y_data]

    return X_data, Y_data


def embedding_prepare(layer, X_data, Y_data):
    """Prepare dataset for Embedding layer object.

    :param layer: An instance of the :class:`epynn.embedding.models.Embedding`
    :type layer: :class:`epynn.embedding.models.Embedding`

    :param X_data: Set of sample features.
    :type encode: list[list] or :class:`numpy.ndarray`

    :param Y_data: Set of samples label.
    :type encode: list[list[int] or int] or :class:`numpy.ndarray`

    :return: All training, validation and testing sets along with batched training set
    :rtype: tuple[:class:`epynn.commons.models.dataSet`]
    """
    # Embedding parameters
    se_dataset = layer.se_dataset

    # Pair-wise features-label list
    dataset = list(zip(X_data, Y_data))

    # Split and separate features and label
    dtrain, dval, dtest = split_dataset(dataset, se_dataset)

    X_train, Y_train = zip(*dtrain)
    X_val, Y_val = zip(*dval) if dval else [(), ()]
    X_test, Y_test = zip(*dtest) if dtest else [(), ()]

    # Instantiate dataSet objects
    dtrain = dataSet(X_data=X_train, Y_data=Y_train, name='dtrain')
    dval = dataSet(X_data=X_val, Y_data=Y_val, name='dval')
    dtest = dataSet(X_data=X_test, Y_data=Y_test, name='dtest')

    embedded_data = (dtrain, dval, dtest)

    return embedded_data


def split_dataset(dataset, se_dataset):
    """Split dataset in training, testing and validation sets.

    :param dataset: Dataset containing sample features and label
    :type dataset: tuple[list or :class:`numpy.ndarray`]

    :param se_dataset: Settings for sets preparation
    :type se_dataset: dict[str: int]

    :return: Training, testing and validation sets.
    :rtype: tuple[list]
    """
    # Retrieve relative sizes
    dtrain_relative = se_dataset['dtrain_relative']
    dval_relative = se_dataset['dval_relative']
    dtest_relative = se_dataset['dtest_relative']

    # Compute absolute sizes with respect to full dataset
    sum_relative = sum([dtrain_relative, dval_relative, dtest_relative])

    dtrain_length = round(dtrain_relative / sum_relative * len(dataset))
    dval_length = round(dval_relative / sum_relative * len(dataset))
    dtest_length = round(dtest_relative / sum_relative * len(dataset))

    # Slice full dataset
    dtrain = dataset[:dtrain_length]
    dval = dataset[dtrain_length:dtrain_length + dval_length]
    dtest = dataset[dtrain_length + dval_length:]

    return dtrain, dval, dtest


def mini_batches(layer):
    """Shuffle and divide dataset in batches for each training epoch.

    :param layer: An instance of the :class:`epynn.embedding.models.Embedding`
    :type layer: :class:`epynn.embedding.models.Embedding`

    :return: Batches made from dataset with respect to batch_size
    :rtype: list[Object]
    """
    # Retrieve training set and make pair-wise features-label dataset
    dtrain_zip = layer.dtrain_zip

    batch_size = layer.se_dataset['batch_size']

    # Shuffle dataset
    if hasattr(layer, 'np_rng'):
        warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning)
        layer.np_rng.shuffle(dtrain_zip)
    else:
        np.random.shuffle(dtrain_zip)

    # Compute number of batches w.r.t. batch_size
    if not batch_size:
        batch_size = len(dtrain_zip)

    n_batch = len(dtrain_zip) // batch_size

    if not n_batch:
        n_batch = 1

    # Slice to make sure split will result in equal division
    dtrain_zip = dtrain_zip[: n_batch * batch_size]

    X_train, Y_train = zip(*dtrain_zip)

    X_train = np.split(np.array(X_train), n_batch, axis=0)
    Y_train = np.split(np.array(Y_train), n_batch, axis=0)

    # Set into dataSet object
    batch_dtrain = [dataSet(X_data=X_batch, Y_data=Y_batch, name=str(i))
                    for i, (X_batch, Y_batch) in enumerate(zip(X_train, Y_train))]

    return batch_dtrain



# EpyNN/epynn/commons/library.py
# Standard library imports
import pathlib
import pickle
import shutil
import glob
import os
 
# Related third party imports
import numpy as np

# Local application/library specific imports
from epynn.commons.logs import process_logs


def read_pickle(f):
    """Read pickle binary file.

    :param f: Filename.
    :type f: str

    :return: File content.
    :rtype: Object
    """
    with open(f, 'rb') as msg:
        c = pickle.load(msg)

    return c


def read_file(f):
    """Read text file.

    :param f: Filename.
    :type f: str

    :return: File content.
    :rtype: str
    """
    with open(f, 'r') as msg:
        c = msg.read()

    return c


def write_pickle(f, c):
    """Write pickle binary file.

    :param f: Filename.
    :type f: str

    :param c: Content to write.
    :type c: Object
    """
    with open(f, 'wb') as msg:
        pickle.dump(c,msg)

    return None


def configure_directory(clear=False):
    """Configure working directory.

    :param clear: Remove and make directories, defaults to False.
    :type clear: bool, optional
    """
    # Set paths for defaults directories
    datasets_path = os.path.join(os.getcwd(), 'datasets')
    models_path = os.path.join(os.getcwd(), 'models')
    plots_path = os.path.join(os.getcwd(), 'plots')

    # Iterate over directory paths
    for path in [datasets_path, models_path, plots_path]:

        # If clear set to True, remove directories
        if clear and os.path.exists(path):
            shutil.rmtree(path)
            process_logs('Remove: '+path, level=2)

        # Create directory if not existing
        if not os.path.exists(path):
            os.mkdir(path)
            process_logs('Make: '+path, level=1)

    return None


def write_model(model, model_path=None):
    """Write EpyNN model on disk.

    :param model: An instance of EpyNN network object.
    :type model: :class:`epynn.network.models.EpyNN`

    :param model_path: Where to write model, defaults to `None` which sets path in `models` directory.
    :type model_path: str or NoneType, optional
    """
    data = {
                'model': model,
            }

    if model_path:
        # If model_path not set to None, pass on user-defined path
        pass
    else:
        # Set default location and name to write model on disk
        model_path = os.path.join(os.getcwd(), 'models', model.uname)
        model_path = model_path+'.pickle'

    # Write model with pickle
    write_pickle(model_path, data)
    process_logs('Make: ' + model_path, level=1)

    return None


def read_model(model_path=None):
    """Read EpyNN model from disk.

    :param model_path: Where to read model from, defaults to `None` which reads the last saved model in `models` directory.
    :type model_path: str or NoneType, optional
    """
    if model_path:
        # If model_path not set to None, pass on user-defined path
        pass
    else:
        # Set default location and name to read the model from
        models_path = os.path.join(os.getcwd(), 'models', '*')
        model_path = max(glob.glob(models_path), key=os.path.getctime)

    model = read_pickle(model_path)['model']

    return model


def settings_verification():
    """Import default :class:`epynn.settings.se_hPars` if not present in working directory.
    """
    # Absolute path of epynn directory
    init_path = str(pathlib.Path(__file__).parent.parent.absolute())

    # Copy defaults settings in working directory if not present
    if not os.path.exists('settings.py'):
        se_default_path = os.path.join(init_path, 'settings.py')
        shutil.copy(se_default_path, 'settings.py')

    return None



# EpyNN/epynn/commons/logs.py
# Standard library imports
import traceback
import json
import time
import sys
 
# Related third party imports
from pygments import highlight, lexers, formatters
from pygments.formatters import TerminalTrueColorFormatter
from pygments.lexers import get_lexer_by_name
from termcolor import cprint, colored
from texttable import Texttable
from pygments import highlight
from tabulate import tabulate
import numpy as np


def headers_logs(model, colors):
    """Generate headers to log epochs, learning rates, training metrics, costs and experiment name.

    :param model: An instance of EpyNN network object.
    :type model: :class:`epynn.network.models.EpyNN`

    :param colors: Colors that may be used to print logs.
    :type colors: list[str]

    :return: Headers with respect to training setup.
    :rtype: list[str]
    """
    metrics = model.metrics                   # Contains metrics + cost computed along training epochs
    len_dsets = len(model.embedding.dsets)    # Number of active datasets (dtrain, dtest, dval or dtrain, dtest or dtrain)

    # Initialize headers list with first field: epochs
    headers = [
        colored('epoch', 'white', attrs=['bold']) + '\n',
    ]

    for layer in model.layers:

        if layer.p != {}:
            # If layer is trainable, then add header for learning rate.
            headers.append(
                colored('lrate', colors[0], attrs=['bold'])
                + '\n'
                + colored(layer.name, colors[0], attrs=[])
            )

    # Iterate over metrics/costs - We need to adapt headers with respect to the number of active datasets
    for i, s in enumerate(list(metrics.keys())):

        i = (i+1) % len(colors)

        # There is at least one non-empty training set
        headers.append(
            colored('%s' % s, colors[i], attrs=['bold'])
            + '\n'
            + colored('dtrain', colors[i], attrs=['bold'])
        )

        # There may be one non-empty validation set
        if len_dsets == 2:
            headers.append('\n' + colored('dval', colors[i], attrs=['bold']))

        # There may be one non-empty validation and testing set
        if len_dsets == 3:
            headers.append('\n' + colored('dtest', colors[i], attrs=['bold']))

    headers.append(colored('Experiment', 'white', attrs=[]) + '\n')

    return headers


def current_logs(model, colors):
    """Build logs with respect to headers for current epoch which includes epoch, learning rates, training metrics, costs and experiment name.

    :param model: An instance of EpyNN network object.
    :type model: :class:`epynn.network.models.EpyNN`

    :param colors: Colors that may be used to print logs.
    :type colors: list[str]

    :return: Logs for current epoch.
    :rtype: list[str]
    """
    metrics = model.metrics          # Contains metrics + cost computed along training epochs
    dsets = model.embedding.dsets    # Active datasets (dtrain, dval, dtest or dtrain, dval or dtrain)

    log = []

    log.append(colored(model.e, 'white', attrs=['bold']))    # Current epoch

    for layer in model.layers:
        # If layer is trainable, append current learning rate
        if layer.p != {}:
            log.append(colored("{:.2e}".format(layer.lrate[model.e]), 'white', attrs=['bold']))

    # Iterate over metrics/costs
    for i, s in enumerate(metrics.keys()):

        i = (i+1) % len(colors)

        # Iterate over each active dataset
        for k in range(len(dsets)):

            m = round(metrics[s][k][-1], 3)

            log.append(colored('%.3f' % m, colors[i], attrs=['bold']))

    log.append(colored(model.uname,'white',attrs=[]))

    return log


def initialize_logs_print(model):
    """Print model initialization logs which include information about datasets, model architecture and shapes as well as layers hyperparameters.

    :param model: An instance of EpyNN network object.
    :type model: :class:`epynn.network.models.EpyNN`
    """
    cprint ('----------------------- %s -------------------------\n' % model.uname, attrs=['bold'], end='\n\n')


    cprint ('-------------------------------- Datasets ------------------------------------\n',attrs=['bold'])

    print (model.init_logs[0].draw(), end='\n\n')
    print (model.init_logs[1].draw(), end='\n\n')


    cprint ('----------------------- Model Architecture -------------------------\n',attrs=['bold'])

    print (model.init_logs[2].draw(), end='\n\n')


    cprint ('------------------------------------------- Layers ---------------------------------------------\n',attrs=['bold'])

    print (model.init_logs[3].draw(), end='\n\n')
    print (model.init_logs[4].draw(), end='\n\n')


    cprint ('----------------------- %s -------------------------\n' % model.uname, attrs=['bold'], end='\n\n')

    return None


def network_logs(network):
    """Build tabular logs of current network architecture and shapes.

    :param network: Attribute of an instance of EpyNN network object.
    :type network: dict[str, dict[str, str or tuple[int]]]

    :return: Logs for network architecture and shapes.
    :rtype: :class:`texttable.Texttable`
    """
    # List of documented network features
    headers = [
        'ID',
        'Layer',
        'Dimensions',
        'Activation',
        'FW_Shapes',
        'BW_Shapes',
    ]

    logs = Texttable()

    logs.add_row(headers)

    # Iterate over values (layers) in network dictionnary
    for i, layer in enumerate(network.values()):

        log = []

        log.append(str(i))            # Layer index
        log.append(layer['Layer'])    # Layer name

        for key in ['Dimensions', 'Activation', 'FW_Shapes', 'BW_Shapes']:
            log.append('\n'.join([k + ': ' + str(v) for k,v in layer[key].items()]))

        logs.add_row(log)

    logs.set_max_width(0)

    return logs


def layers_lrate_logs(layers):
    """Build tabular logs for layers hyperparameters related to learning rate.

    :param layers: Attribute of an instance of EpyNN network object.
    :type layers: list[Object]

    :return: Logs for layers hyperparameters related to learning rate.
    :rtype: :class:`texttable.Texttable`
    """
    # List of documented layers hyperparameters related to learning rate
    headers = [
        'Layer',
        'epochs',
        'schedule',
        'decay_k',
        'cycle\nepochs\n',
        'cycle\ndescent',
        'cycle\nnumber',
        'learning\nrate\n(start)',
        'learning\nrate\n(end)',
        'end\n(%)',
    ]

    logs = Texttable()

    logs.add_row(headers)

    for layer in layers:

        log = []

        se_hPars = layer.se_hPars    # Local hyperparameters for layer

        lr_ori = layer.lrate[0]      # Learning rate epoch 0
        lr_end = layer.lrate[-1]     # Learning rate last epoch

        pc_end = round(lr_end / lr_ori * 100, 3) if lr_ori != 0 else 0

        log.append(layer.name)
        log.append(se_hPars['epochs'])
        log.append(se_hPars['schedule'])
        log.append(se_hPars['decay_k'])
        log.append(se_hPars['cycle_epochs'])
        log.append(se_hPars['cycle_descent'])
        log.append(se_hPars['cycle_number'])
        log.append("{:.2e}".format(lr_ori))
        log.append("{:.2e}".format(lr_end))
        log.append(pc_end)

        # Document only if layer is trainable
        if layer.p != {}:
            logs.add_row(log)

    logs.set_max_width(0)

    return logs


def layers_others_logs(layers):
    """Build tabular logs for layers hyperparameters related to activation functions.

    :param layers: Attribute of an instance of EpyNN network object.
    :type layers: list[Object]

    :return: Logs for layers hyperparameters related to activation functions.
    :rtype: :class:`texttable.Texttable`
    """
    # List of documented layers hyperparameters related to activation functions
    headers = [
        'Layer',
        'LRELU\nalpha',
        'ELU\nalpha',
        'softmax\ntemperature',
    ]

    logs = Texttable()

    logs.add_row(headers)

    for layer in layers:

        se_hPars = layer.se_hPars    # Local hyperparameters for layer

        log = []

        log.append(layer.name)
        log.append(se_hPars['LRELU_alpha'])
        log.append(se_hPars['ELU_alpha'])
        log.append(se_hPars['softmax_temperature'])

        # Document only if layer is trainable
        if layer.p != {}:
            logs.add_row(log)

    logs.set_max_width(0)

    return logs


def dsets_samples_logs(dsets, se_dataset):
    """Build tabular logs describing datasets.

    :param dsets: Attribute of an instance of embedding layer object. Contains active (non-empty) sets.
    :type dsets: list[:class:`epynn.commons.models.dataSet`]

    :param se_dataset: Attribute of an instance of embedding layer object.
    :type se_dataset: dict[str, int or bool]

    :return: Logs describing datasets.
    :rtype: :class:`texttable.Texttable`
    """
    # List of dataset descriptors.
    headers = [
        'dtrain\n',
        'dval\n',
        'dtest\n',
        'batch\nsize',
    ]

    logs = Texttable()

    logs.add_row(headers)

    batch_size = se_dataset['batch_size']

    log = []

    # Active datasets
    for dset in dsets:
        log.append(len(dset.ids))

    # Fill for inactive datasets
    for i in range(3 - len(dsets)):
        log.append('None')

    log.append(batch_size)

    logs.add_row(log)

    logs.set_max_width(0)

    return logs


def dsets_labels_logs(dsets):
    """Build tabular logs describing datasets Y dimension.

    :param dsets: Attribute of an instance of embedding layer object. Contains active (non-empty) sets.
    :type dsets: list[:class:`epynn.commons.models.dataSet`]

    :return: Logs describing datasets Y dimension.
    :rtype: :class:`texttable.Texttable`
    """
    # List of dataset descriptors for Y dimension.
    headers = [
        'N_LABELS',
        'dtrain\n',
        'dval\n',
        'dtest\n',
    ]

    logs = Texttable()

    logs.add_row(headers)

    log = []

    log.append(len(dsets[0].b.keys()))    # Number of classes

    # Active datasets
    for dset in dsets:
        log.append('\n'.join([str(k) + ': ' + str(v) for k,v in sorted(dset.b.items())]))

    # Fill for inactive datasets
    for i in range(3 - len(dsets)):
        log.append('None')

    logs.add_row(log)

    return logs


def set_highlighted_excepthook():
    """Lexer to pretty print tracebacks.
    """
    # Get lexer from pigmentize/minted
    lexer = get_lexer_by_name('pytb' if sys.version_info.major < 3 else 'py3tb')

    # Colorscheme
    formatter = TerminalTrueColorFormatter(bg='dark', style='fruity')

    # Callback function
    def myexcepthook(type, value, tb):
        tbtext = ''.join(traceback.format_exception(type, value, tb))
        sys.stderr.write(highlight(tbtext, lexer, formatter))

        return None

    sys.excepthook = myexcepthook    # This erase the standard excepthook with the callback

    return None


def process_logs(msg, level=0):
    """Pretty print of EpyNN events.

    :param msg: Message to print on terminal.
    :type msg: str

    :param level: Set color for print, defaults to 0 which renders white.
    :type level: int, optional
    """
    colors = [
        'white',
        'green',
        'red',
        'magenta',
        'cyan',
        'yellow',
        'blue',
        'grey',
    ]

    cprint(msg, colors[level],  attrs=['bold'])

    return None


def start_counter(timeout=3):
    """Timeout between print of initialization logs and beginning of run.

    :param timeout: Time in seconds, defaults to 3.
    :type timeout: int, optional
    """
    for i in reversed(range(timeout + 1)):
        cprint('Start in %ss' % str(i), attrs=['bold'], end='\r')
        time.sleep(1)

    return None


def pretty_json(network):
    """Pretty json print for traceback during model initialization.

    :param network: Attribute of an instance of EpyNN network object.
    :type network: dict[str, dict[str, str or tuple[int]]]

    :return: Formatted input.
    :rtype: json
    """
    # Convert dict to json
    network = json.dumps(network, sort_keys=False, indent=4)

    # Format for pretty print
    network = highlight(network, lexers.JsonLexer(), formatters.TerminalFormatter())

    return network



# EpyNN/epynn/lstm/parameters.py
# Related third party imports
import numpy as np


def lstm_compute_shapes(layer, A):
    """Compute forward shapes and dimensions from input for layer.
    """
    X = A    # Input of current layer

    layer.fs['X'] = X.shape    # (m, s, e)

    layer.d['m'] = layer.fs['X'][0]    # Number of samples (m)
    layer.d['s'] = layer.fs['X'][1]    # Steps in sequence (s)
    layer.d['e'] = layer.fs['X'][2]    # Elements per step (e)

    # Parameter Shapes             Unit cells (u)
    eu = (layer.d['e'], layer.d['u'])    # (v, u)
    uu = (layer.d['u'], layer.d['u'])    # (u, u)
    u1 = (1, layer.d['u'])               # (1, u)
    # Forget gate    Input gate       Candidate        Output gate
    layer.fs['Uf'] = layer.fs['Ui'] = layer.fs['Ug'] = layer.fs['Uo'] = eu
    layer.fs['Vf'] = layer.fs['Vi'] = layer.fs['Vg'] = layer.fs['Vo'] = uu
    layer.fs['bf'] = layer.fs['bi'] = layer.fs['bg'] = layer.fs['bo'] = u1

    # Shape of hidden (h) and memory (C) state with respect to steps (s)
    layer.fs['h'] = layer.fs['C'] = (layer.d['m'], layer.d['s'], layer.d['u'])

    return None


def lstm_initialize_parameters(layer):
    """Initialize trainable parameters from shapes for layer.
    """
    # For linear activation of forget gate (f_)
    layer.p['Uf'] = layer.initialization(layer.fs['Uf'], rng=layer.np_rng)
    layer.p['Vf'] = layer.initialization(layer.fs['Vf'], rng=layer.np_rng)
    layer.p['bf'] = np.zeros(layer.fs['bf']) # dot(X, U) + dot(hp, V) + b

    # For linear activation of input gate (i_)
    layer.p['Ui'] = layer.initialization(layer.fs['Ui'], rng=layer.np_rng)
    layer.p['Vi'] = layer.initialization(layer.fs['Vi'], rng=layer.np_rng)
    layer.p['bi'] = np.zeros(layer.fs['bi']) # dot(X, U) + dot(hp, V) + b

    # For linear activation of candidate (g_)
    layer.p['Ug'] = layer.initialization(layer.fs['Ug'], rng=layer.np_rng)
    layer.p['Vg'] = layer.initialization(layer.fs['Vg'], rng=layer.np_rng)
    layer.p['bg'] = np.zeros(layer.fs['bg']) # dot(X, U) + dot(hp, V) + b

    # For linear activation of output gate (o_)
    layer.p['Uo'] = layer.initialization(layer.fs['Uo'], rng=layer.np_rng)
    layer.p['Vo'] = layer.initialization(layer.fs['Vo'], rng=layer.np_rng)
    layer.p['bo'] = np.zeros(layer.fs['bo']) # dot(X, U) + dot(hp, V) + b

    return None


def lstm_compute_gradients(layer):
    """Compute gradients with respect to weight and bias for layer.
    """
    # Gradients initialization with respect to parameters
    for parameter in layer.p.keys():
        gradient = 'd' + parameter
        layer.g[gradient] = np.zeros_like(layer.p[parameter])

    # Reverse iteration over sequence steps
    for s in reversed(range(layer.d['s'])):

        X = layer.fc['X'][:, s]      # Input for current step
        hp = layer.fc['hp'][:, s]    # Previous hidden state

        # (1) Gradients of the loss with respect to U, V, b
        do_ = layer.bc['do_'][:, s]            # Gradient w.r.t output gate o_
        layer.g['dUo'] += np.dot(X.T, do_)     # (1.1) dL/dUo
        layer.g['dVo'] += np.dot(hp.T, do_)    # (1.2) dL/dVo
        layer.g['dbo'] += np.sum(do_, axis=0)  # (1.3) dL/dbo

        # (2) Gradients of the loss with respect to U, V, b
        dg_ = layer.bc['dg_'][:, s]            # Gradient w.r.t candidate g_
        layer.g['dUg'] += np.dot(X.T, dg_)     # (2.1) dL/dUg
        layer.g['dVg'] += np.dot(hp.T, dg_)    # (2.2) dL/dVg
        layer.g['dbg'] += np.sum(dg_, axis=0)  # (2.3) dL/dbg

        # (3) Gradients of the loss with respect to U, V, b
        di_ = layer.bc['di_'][:, s]            # Gradient w.r.t input gate i_
        layer.g['dUi'] += np.dot(X.T, di_)     # (3.1) dL/dUi
        layer.g['dVi'] += np.dot(hp.T, di_)    # (3.2) dL/dVi
        layer.g['dbi'] += np.sum(di_, axis=0)  # (3.3) dL/dbi

        # (4) Gradients of the loss with respect to U, V, b
        df_ = layer.bc['df_'][:, s]            # Gradient w.r.t forget gate f_
        layer.g['dUf'] += np.dot(X.T, df_)     # (4.1) dL/dUf
        layer.g['dVf'] += np.dot(hp.T, df_)    # (4.2) dL/dVf
        layer.g['dbf'] += np.sum(df_, axis=0)  # (4.3) dL/dbf

    return None


def lstm_update_parameters(layer):
    """Update parameters from gradients for layer.
    """
    for gradient in layer.g.keys():
        parameter = gradient[1:]
        # Update is driven by learning rate and gradients
        layer.p[parameter] -= layer.lrate[layer.e] * layer.g[gradient]

    return None



# EpyNN/epynnlive/dummy_boolean/train.py
# Standard library imports
import random

# Related third party imports
import numpy as np

# Local application/library specific imports
import epynn.initialize
from epynn.commons.library import (
    configure_directory,
    read_model,
)
from epynn.network.models import EpyNN
from epynn.embedding.models import Embedding
from epynn.dense.models import Dense
from prepare_dataset import prepare_dataset


########################## CONFIGURE ##########################
random.seed(1)

np.set_printoptions(threshold=10)

np.seterr(all='warn')

configure_directory(clear=True)    # This is a dummy example


############################ DATASET ##########################
X_features, Y_label = prepare_dataset(N_SAMPLES=50)


####################### BUILD AND TRAIN MODEL #################

embedding = Embedding(X_data=X_features,
                      Y_data=Y_label,
                      relative_size=(2, 1, 0))


### Feed-Forward

# Model
name = 'Perceptron_Dense-1-sigmoid'

dense = Dense()

model = EpyNN(layers=[embedding, dense], name=name)

model.train(epochs=100)

model.plot(path=False)


### Write/read model

model.write()
# model.write(path=/your/custom/path)

model = read_model()
# model = read_model(path=/your/custom/path)


### Predict

X_features, _ = prepare_dataset(N_SAMPLES=10)

dset = model.predict(X_features)

for n, pred, probs, features in zip(dset.ids, dset.P, dset.A, dset.X):
    print(n, pred, probs, features)



# EpyNN/epynn/commons/plot.py
# Standard library imports
import os

# Related third party imports
from matplotlib import pyplot as plt

# Local application/library specific imports
from epynn.commons.logs import process_logs
 

def pyplot_metrics(model, path):
    """Plot metrics/costs from training with matplotlib.

    :param model: An instance of EpyNN network object.
    :type model: :class:`epynn.meta.models.EpyNN`

    :param path: Write matplotlib plot.
    :type path: bool or NoneType
    """
    plt.figure()

    metrics = model.metrics    # Contains metrics and cost

    # Iterate over metrics/costs
    for s in metrics.keys():

        # Iterate over active datasets
        for k, dset in enumerate(model.embedding.dsets):

            dname = dset.name

            x = [x for x in range(len(metrics[s][k]))]    # X range

            y = metrics[s][k]    # Y values from metrics[idx_dataset][idx_metrics]

            plt.plot(x, y, label=dname + ' ' + s)

    plt.legend()

    plt.xlabel('Epoch')
    plt.ylabel('Value')

    plt.title(model.uname)

    plt.show()

    # If path sets to None, set to defaults - Note path can be set to False, which makes it print only
    if path == None:
        path = 'plots'
        plot_path = os.path.join(os.getcwd(), path, model.uname)  + '.png'

    if path:
        plt.savefig(plot_path)
        process_logs('Make: ' + plot_path, level=1)

    plt.close()

    return None



# EpyNN/epynn/network/backward.py


def model_backward(model, dA):
    """Backward propagate error gradients from output to input layer.
    """
    # By convention
    dX = dA

    # Iterate over reversed layers
    for layer in reversed(model.layers):

        # Layer returns dL/dX (dX) - layer.bs, layer.bc
        dX = layer.backward(dX)

        # Update values in layer.g
        layer.compute_gradients()

        # Update values in layer.p
        layer.update_parameters()

    return None



# EpyNN/epynnlive/dummy_boolean/settings.py


# HYPERPARAMETERS SETTINGS
se_hPars = {
    # Schedule learning rate
    'learning_rate': 0.1,
    'schedule_mode': 'steady',
    'decay_k': 0.001,
    'cycling_n': 1,
    'descent_d': 1,
    # Tune activation function
    'ELU_alpha': 0.01,
    'LRELU_alpha': 0.01,
    'softmax_temperature': 1,
}
"""Hyperparameters dictionary settings.

Set hyperparameters for model and layer.
"""



# EpyNN/epynn/network/evaluate.py
# Related third party imports
import numpy as np

# Local application/library specific imports
from epynn.commons.loss import loss_functions
from epynn.commons.metrics import metrics_functions


def model_evaluate(model):
    """Compute metrics including cost for model.

    Will evaluate training, testing and validation sets against metrics set in model.se_config.

    :param model: An instance of EpyNN network.
    :type model: :class:`epynn.network.models.EpyNN`
    """

    # Callback functions for metrics and loss
    metrics = metrics_functions()
    metrics.update(loss_functions())

    dsets = model.embedding.dsets

    # Iterate over dsets [dtrain, dval, dtest]
    for k, dset in enumerate(dsets):

        # Check if one-hot encoding
        encoded = (dset.Y.shape[1] > 1)

        # Output probs
        dset.A = model.forward(dset.X)

        # Decisions
        dset.P = np.argmax(dset.A, axis=1) if encoded else np.around(dset.A)

        # Iterate over selected metrics
        for s in model.metrics.keys():

            m = metrics[s](dset.Y, dset.A)

            # Metrics such as precision/recall returned as scalar
            if m.ndim == 0:
                pass
            # Others returned as per-sample 1D array
            else:
                m = np.mean(m)  # To scalar

            # Save value for metrics (s) for dset (k)
            model.metrics[s][k].append(m)

    return None


def batch_evaluate(model, Y, A):
    """Compute metrics for current batch.

    Will evaluate current batch against accuracy and training loss.

    :param model: An instance of EpyNN network.
    :type model: :class:`epynn.network.models.EpyNN`

    :param Y: True labels for batch samples.
    :type Y: :class:`numpy.ndarray`

    :param A: Output of forward propagation for batch.
    :type A: :class:`numpy.ndarray`
    """
    metrics = metrics_functions()

    # Per sample 1D array to scalar
    accuracy = np.mean(metrics['accuracy'](Y, A))

    # Per sample 1D array to scalar
    cost = np.mean(model.training_loss(Y, A))

    return accuracy, cost



# EpyNN/epynn/dropout/paremeters.py
# Related third party imports
import numpy as np


def dropout_compute_shapes(layer, A):
    """Compute forward shapes and dimensions from input for layer.
    """
    X = A    # Input of current layer

    layer.fs['X'] = X.shape    # (m, .. )

    layer.d['m'] = layer.fs['X'][0]          # Number of samples  (m)
    layer.d['n'] = X.size // layer.d['m']    # Number of features (n)

    # Shape for dropout mask
    layer.fs['D'] = [1 if ax in layer.d['a'] else layer.fs['X'][ax]
                     for ax in range(X.ndim)]

    return None


def dropout_initialize_parameters(layer):
    """Initialize trainable parameters from shapes for layer.
    """
    # No parameters to initialize for Dropout layer

    return None


def dropout_compute_gradients(layer):
    """Compute gradients with respect to weight and bias for layer.
    """
    # No gradients to update for Dropout layer

    return None


def dropout_update_parameters(layer):
    """Update parameters from gradients for layer.
    """
    # No parameters to update for Dropout layer

    return None



# EpyNN/epynn/network/initialize.py
# Standard library imports
import sys

# Related third party imports
from termcolor import cprint
import numpy as np


def model_initialize(model, params=True, end='\n'):
    """Initialize EpyNN network.

    :param model: An instance of EpyNN network.
    :type model: :class:`epynn.network.models.EpyNN`

    :param params: Layer parameters initialization, defaults to `True`.
    :type params: bool, optional

    :param end: Wether to print every line for steps or overwrite, default to `\\n`.
    :type end: str in ['\\n', '\\r']

    :raises Exception: If any layer other than Dense was provided with softmax activation. See :func:`epynn.maths.softmax`.
    """
    # Retrieve sample batch
    model.embedding.training_batches(init=True)
    batch_dtrain = model.embedding.batch_dtrain
    batch = batch_dtrain[0]
    A = batch.X  # Features
    Y = batch.Y  # Labels

    cprint('{: <100}'.format('--- EpyNN Check --- '), attrs=['bold'], end=end)

    # Iterate over layers
    for layer in model.layers:

        # Layer instance attributes
        layer.check = False
        layer.name = layer.__class__.__name__

        cprint('Layer: ' + layer.name, attrs=['bold'], end=end)

        # Store layer information in model summary
        model.network[id(layer)]['Layer'] = layer.name
        model.network[id(layer)]['Activation'] = layer.activation
        model.network[id(layer)]['Dimensions'] = layer.d

        # Dense uses epynn.maths.hadamard to handle softmax derivative
        if 'softmax' in layer.activation.values() and layer.name != 'Dense':
            raise Exception('Softmax can not be used with %s, only with Dense' % layer.name)

        # Test layer.compute_shapes() method
        cprint('compute_shapes: ' + layer.name, 'green', attrs=['bold'], end=end)
        layer.compute_shapes(A)

        # Store forward shapes in model summary
        model.network[id(layer)]['FW_Shapes'] = layer.fs

        # Initialize trainable parameters
        if params:
            cprint('initialize_parameters: ' + layer.name, 'green', attrs=['bold'], end=end)
            layer.initialize_parameters()

        # Test layer.forward() method
        cprint('forward: ' + layer.name, 'green', attrs=['bold'], end=end)
        A = layer.forward(A)

        # Output shape
        print('shape:', layer.fs['A'], end=end)

        # Store updated forward shapes in model summary
        model.network[id(layer)]['FW_Shapes'] = layer.fs

        # Clear check
        delattr(layer, 'check')

    # Compute derivative of loss function
    dX = dA = model.training_loss(Y, A, deriv=True)

    # Iterate over reversed layers
    for layer in reversed(model.layers):

        # Set check attribute for layer
        layer.check = False

        cprint('Layer: ' + layer.name, attrs=['bold'], end=end)

        # Test layer.backward() method
        cprint('backward: ' + layer.name, 'cyan', attrs=['bold'], end=end)
        dX = layer.backward(dX)

        # Output shape
        print('shape:', layer.bs['dX'], end=end)

        # Store backward shapes in model summary
        model.network[id(layer)]['BW_Shapes'] = layer.bs

        # Test layer.compute_gradients() method
        cprint('compute_gradients: ' + layer.name, 'cyan', attrs=['bold'], end=end)
        layer.compute_gradients()

        # Clear check
        delattr(layer, 'check')

    cprint('{: <100}'.format('--- EpyNN Check OK! --- '), attrs=['bold'], end=end)

    # Initialize current epoch to zero
    model.e = 0

    return None


def model_assign_seeds(model):
    """Seed model and layers with independant pseudo-random number generators.

    Model is seeded from user-input. Layers are seeded by incrementing the
    input by one in order to not generate same numbers for all objects

    :param model: An instance of EpyNN network.
    :type model: :class:`epynn.network.models.EpyNN`
    """
    seed = model.seed

    # If seed is not defined, seeding is random
    model.np_rng = np.random.default_rng(seed=seed)

    # Iterate over layers
    for layer in model.layers:

        # If seed is defined
        if seed:
            # We do not want the same seed for every object
            seed += 1

        # Seed layer
        layer.o['seed'] = seed
        layer.np_rng = np.random.default_rng(seed=layer.o['seed'])

    return None


def model_initialize_exceptions(model, trace):
    """Handle error in model initialization and show logs.

    :param model: An instance of EpyNN network.
    :type model: :class:`epynn.network.models.EpyNN`

    :param trace: Traceback of fatal error.
    :type trace: traceback object
    """
    cprint('\n/!\\ Initialization of EpyNN model failed - debug', 'red', attrs=['bold'])

    try:
        # Identify faulty layer
        layer = [layer for layer in model.layers if hasattr(layer, 'check')][0]

        # Update shapes from existing caches
        layer.update_shapes(layer.fc, layer.fs)
        layer.update_shapes(layer.bc, layer.bs)

        # Report debug information for faulty layer
        cprint('%s layer: ' % layer.name, 'red', attrs=['bold'])

        cprint('Known dimensions', 'white', attrs=['bold'])
        print(', '.join([k + ': ' + str(v) for k, v in layer.d.items()]))

        cprint('Known forward shapes', 'green', attrs=['bold'])
        print('\n'.join([k + ': ' + str(v) for k, v in layer.fs.items()]))

        cprint('Known backward shape', 'cyan', attrs=['bold'])
        print('\n'.join([k + ': ' + str(v) for k, v in layer.bs.items()]))

    except:
        pass

    # Report traceback of error and exit program
    cprint('System trace', 'red', attrs=['bold'])
    print(trace)

    sys.exit()



# EpyNN/epynn/settings.py


# HYPERPARAMETERS SETTINGS
se_hPars = {
    # Schedule learning rate
    'learning_rate': 0.1,
    'schedule': 'steady',
    'decay_k': 0,
    'cycle_epochs': 0,
    'cycle_descent': 0,
    # Tune activation function
    'ELU_alpha': 0.01,
    'LRELU_alpha': 0.01,
    'softmax_temperature': 1,
}
"""Hyperparameters dictionary settings.

Set hyperparameters for model and layer.
"""



# EpyNN/epynn/network/train.py


def model_training(model):
    """Perform the training of the Neural Network.

    :param model: An instance of EpyNN network.
    :type model: :class:`epynn.network.models.EpyNN`
    """
    # Iterate over training epochs
    for model.e in range(model.e, model.epochs):

        # Shuffle dtrain and prepare new batches
        model.embedding.training_batches()

        # Iterate over training batches
        for batch in model.embedding.batch_dtrain:

            # Pass through every layer.forward() methods
            A = model.forward(batch.X)

            # Compute derivative of loss
            dA = model.training_loss(batch.Y, A, deriv=True)

            # Pass through every layer.backward() methods
            model.backward(dA)

            # Accuracy and cost for batch
            model.batch_report(batch, A)

        # Selected metrics and costs for dsets
        model.evaluate()

        # Tabular report for dsets
        model.report()

    return None



# EpyNN/epynn/pool/backward.py
# Related third party imports
import numpy as np


def initialize_backward(layer, dX):
    """Backward cache initialization.

    :param layer: An instance of pooling layer.
    :type layer: :class:`epynn.pooling.models.Pooling`

    :param dX: Output of backward propagation from next layer.
    :type dX: :class:`numpy.ndarray`

    :return: Input of backward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    dA = layer.bc['dA'] = dX

    return dA


def pooling_backward(layer, dX):
    """Backward propagate error gradients to previous layer.
    """
    # (1) Initialize cache
    dA = initialize_backward(layer, dX)    # (m, oh, ow, d)

    # (2) Restore pooling block axes
    dZ = dA
    dZ = np.expand_dims(dZ, axis=3)
    dZ = np.expand_dims(dZ, axis=3)
    # (m, oh, ow, d)         ->
    # (m, oh, ow, 1, d)      ->
    # (m, oh, ow, 1, 1, d)

    # (3) Initialize backward output dL/dX
    dX = np.zeros_like(layer.fc['X'])      # (m, h, w, d)

    # Iterate over forward output height
    for oh in range(layer.d['oh']):

        hs = oh * layer.d['sh']
        he = hs + layer.d['ph']

        # Iterate over forward output width
        for ow in range(layer.d['ow']):

            ws = ow * layer.d['sw']
            we = ws + layer.d['pw']

            # (4hw) Retrieve input block
            Xb = layer.fc['Xb'][:, oh, ow, :, :, :]
            # (m, oh, ow, ph, pw, d)  - Xb (array of blocks)
            # (m, ph, pw, d)          - Xb (single block)

            # (5hw) Retrieve pooled value and restore block shape
            Zb = layer.fc['Z'][:, oh:oh+1, ow:ow+1, :]
            Zb = np.repeat(Zb, layer.d['ph'], axis=1)
            Zb = np.repeat(Zb, layer.d['pw'], axis=2)
            # (m, oh, ow, d)    - Z
            # (m,  1,  1, d)    - Zb -> np.repeat(Zb, pw, axis=1)
            # (m, ph,  1, d)         -> np.repeat(Zb, pw, axis=2)
            # (m, ph, pw, d)

            # (6hw) Match pooled value in Zb against Xb
            mask = (Zb == Xb)

            # (7hw) Retrieve gradient w.r.t Z and restore block shape
            dZb = dZ[:, oh, ow, :]
            dZb = np.repeat(dZb, layer.d['ph'], 1)
            dZb = np.repeat(dZb, layer.d['pw'], 2)
            # (m, oh, ow,  1,  1, d) - dZ
            #         (m,  1,  1, d) - dZb -> np.repeat(dZb, ph, axis=1)
            #         (m, ph,  1, d)       -> np.repeat(dZb, pw, axis=2)
            #         (m, ph, pw, d)

            # (8hw) Keep dXb values for coordinates where Zb = Xb (mask)
            dXb = dZb * mask

            # (9hw) Gradient of the loss w.r.t Xb
            dX[:, hs:he, ws:we, :] += dXb
            # (m, ph, pw, d) - dX[:, hs:he, ws:we, :]
            # (m, ph, pw, d) - dXb

    layer.bc['dX'] = dX

    return dX



# EpyNN/epynn/convolution/forward.py
# Related third party imports
import numpy as np

# Local application/library specific imports
from epynn.commons.io import padding


def initialize_forward(layer, A):
    """Forward cache initialization.

    :param layer: An instance of convolution layer.
    :type layer: :class:`epynn.convolution.models.Convolution`

    :param A: Output of forward propagation from previous layer.
    :type A: :class:`numpy.ndarray`

    :return: Input of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`

    :return: Input of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`

    :return: Input blocks of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    X = layer.fc['X'] = padding(A, layer.d['p'])

    return X


def convolution_forward(layer, A):
    """Forward propagate signal to next layer.
    """
    # (1) Initialize cache and pad image
    X = initialize_forward(layer, A)    # (m, h, w, d)

    # (2) Slice input w.r.t. filter size (fh, fw) and strides (sh, sw)
    Xb = np.array([[X[ :, h:h + layer.d['fh'], w:w + layer.d['fw'], :]
                    # Inner loop
                    # (m, h, w, d) ->
                    # (ow, m, h, fw, d)
                    for w in range(layer.d['w'] - layer.d['fw'] + 1)
                    if w % layer.d['sw'] == 0]
                # Outer loop
                # (ow, m, h, fw, d) ->
                # (oh, ow, m, fh, fw, d)
                for h in range(layer.d['h'] - layer.d['fh'] + 1)
                if h % layer.d['sh'] == 0])

    # (3) Bring back m along axis 0
    Xb = np.moveaxis(Xb, 2, 0)
    # (oh, ow, m, fh, fw, d) ->
    # (m, oh, ow, fh, fw, d)

    # (4) Add dimension for filter units (u) on axis 6
    Xb = layer.fc['Xb'] = np.expand_dims(Xb, axis=6)
    # (m, oh, ow, fh, fw, d) ->
    # (m, oh, ow, fh, fw, d, 1)

    # (5.1) Linear activation Xb -> Zb
    Zb = Xb * layer.p['W']
    # (m, oh, ow, fh, fw, d, 1) - Xb
    #            (fh, fw, d, u) - W

    # (5.2) Sum block products
    Z = np.sum(Zb, axis=(5, 4, 3))
    # (m, oh, ow, fh, fw, d, u) - Zb
    # (m, oh, ow, fh, fw, u)    - np.sum(Zb, axis=(5))
    # (m, oh, mw, fh, u)        - np.sum(Zb, axis=(5, 4))
    # (m, oh, ow, u)            - np.sum(Zb, axis=(5, 4, 3))

    # (5.3) Add bias to linear activation product
    Z = layer.fc['Z'] = Z + layer.p['b']

    # (6) Non-linear activation
    A = layer.fc['A'] = layer.activate(Z)

    return A    # To next layer



# EpyNN/epynn/dropout/models.py
# Local application/library specific imports
from epynn.commons.models import Layer
from epynn.dropout.forward import dropout_forward
from epynn.dropout.backward import dropout_backward
from epynn.dropout.parameters import (
    dropout_compute_shapes,
    dropout_initialize_parameters,
    dropout_compute_gradients,
    dropout_update_parameters
)


class Dropout(Layer):
    """
    Definition of a dropout layer prototype.

    :param drop_prob: Probability to drop one data point from previous layer to next layer, defaults to 0.5.
    :type drop_prob: float, optional

    :param axis: Compute and apply dropout mask along defined axis, defaults to all axis.
    :type axis: int or tuple[int], optional
    """

    def __init__(self, drop_prob=0.5, axis=()):
        """Initialize instance variable attributes.
        """
        super().__init__()

        axis = axis if isinstance(axis, tuple) else (axis,)

        self.d['d'] = drop_prob
        self.d['a'] = axis

        self.trainable = False

        return None

    def compute_shapes(self, A):
        """Wrapper for :func:`epynn.dropout.parameters.dropout_compute_shapes()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`
        """
        dropout_compute_shapes(self, A)

        return None

    def initialize_parameters(self):
        """Wrapper for :func:`epynn.dropout.parameters.dropout_initialize_parameters()`.
        """
        dropout_initialize_parameters(self)

        return None

    def forward(self, A):
        """Wrapper for :func:`epynn.dropout.forward.dropout_forward()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`

        :return: Output of forward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        self.compute_shapes(A)
        A = self.fc['A'] = dropout_forward(self, A)
        self.update_shapes(self.fc, self.fs)

        return A

    def backward(self, dX):
        """Wrapper for :func:`epynn.dropout.backward.dropout_backward()`.

        :param dX: Output of backward propagation from next layer.
        :type dX: :class:`numpy.ndarray`

        :return: Output of backward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        dX = dropout_backward(self, dX)
        self.update_shapes(self.bc, self.bs)

        return dX

    def compute_gradients(self):
        """Wrapper for :func:`epynn.dropout.parameters.dropout_compute_gradients()`. Dummy method, there are no gradients to compute in layer.
        """
        dropout_compute_gradients(self)

        return None

    def update_parameters(self):
        """Wrapper for :func:`epynn.dropout.parameters.dropout_update_parameters()`. Dummy method, there are no parameters to update in layer.
        """
        if self.trainable:
            dropout_update_parameters(self)

        return None



# EpyNN/epynnlive/dummy_boolean/prepare_dataset.py
# Standard library imports
import random


def features_boolean(N_FEATURES=11):
    """Generate dummy string features.

    :param N_FEATURES: Number of features, defaults to 11.
    :type N_FEATURES: int

    :return: random boolean features of length N_FEATURES.
    :rtype: list[bool]
    """
    # Random choice True or False for N_FEATURES iterations
    features = [random.choice([True, False]) for j in range(N_FEATURES)]

    return features


def label_features(features):
    """Prepare label associated with features.

    The dummy law is:

    More True = positive.
    More False = negative.

    :param features: random boolean features of length N_FEATURES.
    :type features: list[bool]

    :return: Single-digit label with respect to features.
    :rtype: int
    """
    # Single-digit positive and negative labels
    p_label = 0
    n_label = 1

    # Test if features contains more True (0)
    if features.count(True) > features.count(False):
        label = p_label

    # Test if features contains more False (1)
    elif features.count(True) < features.count(False):
        label = n_label

    return label


def prepare_dataset(N_SAMPLES=100):
    """Prepare a set of dummy boolean sample features and label.

    :param N_SAMPLES: Number of samples to generate, defaults to 100.
    :type N_SAMPLES: int

    :return: Set of sample features.
    :rtype: tuple[list[bool]]

    :return: Set of single-digit sample label.
    :rtype: tuple[int]
    """
    # Initialize X and Y datasets
    X_features = []
    Y_label = []

    # Iterate over N_SAMPLES
    for i in range(N_SAMPLES):

        # Compute random boolean features
        features = features_boolean()

        # Retrieve label associated with features
        label = label_features(features)

        # Append sample features to X_features
        X_features.append(features)

        # Append sample label to Y_label
        Y_label.append(label)

    # Prepare X-Y pairwise dataset
    dataset = list(zip(X_features, Y_label))

    # Shuffle dataset
    random.shuffle(dataset)

    # Separate X-Y pairs
    X_features, Y_label = zip(*dataset)

    return X_features, Y_label



# EpyNN/epynn/template/parameters.py


def template_compute_shapes(layer, A):
    """Compute forward shapes and dimensions from input for layer.
    """
    X = A    # Input of current layer

    layer.fs['X'] = X.shape    # (m, .. )

    layer.d['m'] = layer.fs['X'][0]          # Number of samples (m)
    layer.d['n'] = X.size // layer.d['m']    # Number of features (n)

    return None


def template_initialize_parameters(layer):
    """Initialize parameters from shapes for layer.
    """
    # No parameters to initialize for Template layer

    return None


def template_compute_gradients(layer):
    """Compute gradients with respect to weight and bias for layer.
    """
    # No gradients to compute for Template layer

    return None


def template_update_parameters(layer):
    """Update parameters from gradients for layer.
    """
    # No parameters to update for Template layer

    return None



# EpyNN/epynn/flatten/forward.py
# Related third party imports
import numpy as np


def initialize_forward(layer, A):
    """Forward cache initialization.

    :param layer: An instance of flatten layer.
    :type layer: :class:`epynn.flatten.models.Flatten`

    :param A: Output of forward propagation from previous layer.
    :type A: :class:`numpy.ndarray`

    :return: Input of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    X = layer.fc['X'] = A

    return X


def flatten_forward(layer,A):
    """Forward propagate signal to next layer.
    """
    # (1) Initialize cache
    X = initialize_forward(layer, A)

    # (2) Reshape  (m, ...) -> (m, n)
    A = layer.fc['A'] = np.reshape(X, layer.fs['A'])

    return A    # To next layer



# EpyNN/epynn/rnn/forward.py
# Related third party imports
import numpy as np


def initialize_forward(layer, A):
    """Forward cache initialization.

    :param layer: An instance of RNN layer.
    :type layer: :class:`epynn.rnn.models.RNN`

    :param A: Output of forward propagation from previous layer.
    :type A: :class:`numpy.ndarray`

    :return: Input of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`

    :return: Previous hidden state initialized with zeros.
    :rtype: :class:`numpy.ndarray`
    """
    X = layer.fc['X'] = A

    cache_keys = ['h_', 'h', 'hp']
    layer.fc.update({k: np.zeros(layer.fs['h']) for k in cache_keys})

    h = layer.fc['h'][:, 0]    # Hidden state

    return X, h


def rnn_forward(layer, A):
    """Forward propagate signal to next layer.
    """
    # (1) Initialize cache and hidden state
    X, h = initialize_forward(layer, A)

    # Iterate over sequence steps
    for s in range(layer.d['s']):

        # (2s) Slice sequence (m, s, e) with respect to step
        X = layer.fc['X'][:, s]

        # (3s) Retrieve previous hidden state
        hp = layer.fc['hp'][:, s] = h

        # (4s) Activate current hidden state
        h_ = layer.fc['h_'][:, s] = (
            np.dot(X, layer.p['U'])
            + np.dot(hp, layer.p['V'])
            + layer.p['b']
        )   # (4.1s) Linear

        h = layer.fc['h'][:, s] = layer.activate(h_)   # (4.2s) Non-linear

    # Return the last hidden state or the full sequence
    A = layer.fc['h'] if layer.sequences else layer.fc['h'][:, -1]

    return A   # To next layer



# EpyNN/epynn/pooling/forward.py
# Related third party imports
import numpy as np


def initialize_forward(layer, A):
    """Forward cache initialization.

    :param layer: An instance of pooling layer.
    :type layer: :class:`epynn.pooling.models.Pooling`

    :param A: Output of forward propagation from previous layer.
    :type A: :class:`numpy.ndarray`

    :return: Input of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`

    :return: Input of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`

    :return: Input blocks of forward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    X = layer.fc['X'] = A

    return X


def pooling_forward(layer, A):
    """Forward propagate signal to next layer.
    """
    # (1) Initialize cache
    X = initialize_forward(layer, A)

    # (2) Slice input w.r.t. pool size (ph, pw) and strides (sh, sw)
    Xb = np.array([[X[ :, h:h + layer.d['ph'], w:w + layer.d['pw'], :]
                    # Inner loop
                    # (m, h, w, d) ->
                    # (ow, m, h, pw, d)
                    for w in range(layer.d['w'] - layer.d['pw'] + 1)
                    if w % layer.d['sw'] == 0]
                # Outer loop
                # (ow, m, h, pw, d) ->
                # (oh, ow, m, ph, pw, d)
                for h in range(layer.d['h'] - layer.d['ph'] + 1)
                if h % layer.d['sh'] == 0])

    # (3) Bring back m along axis 0
    Xb = layer.fc['Xb'] = np.moveaxis(Xb, 2, 0)
    # (oh, ow, m, ph, pw, d) ->
    # (m, oh, ow, ph, pw, d)

    # (4) Apply pooling operation on blocks
    Z = layer.pool(Xb, axis=(4, 3))
    # (m, oh, ow, ph, pw, d) - Xb
    # (m, oh, ow, ph, d)     - layer.pool(Xb, axis=4)
    # (m, oh, ow, d)         - layer.pool(Xb, axis=(4, 3))

    A = layer.fc['A'] = layer.fc['Z'] = Z

    return A    # To next layer



# EpyNN/epynn/network/hyperparameters.py
# Related third party imports
import numpy as np

# Local application/library specific imports
from epynn.commons.schedule import schedule_functions


def model_hyperparameters(model):
    """Set hyperparameters for each layer in model.

    :param model: An instance of EpyNN network.
    :type model: :class:`epynn.network.models.EpyNN`
    """
    # Iterate over layers
    for layer in model.layers:

        # If no se_hPars provided for layer, then assign se_hPars from model
        if not layer.se_hPars:
            layer.se_hPars = model.se_hPars
        else:
            pass

    return None


def model_learning_rate(model):
    """Schedule learning rate for each layer in model.

    :param model: An instance of EpyNN network.
    :type model: :class:`epynn.network.models.EpyNN`
    """
    # Iterate over layers
    for layer in model.layers:
        layer.se_hPars, layer.lrate = schedule_lrate(layer.se_hPars, model.epochs)

    return None


def schedule_lrate(se_hPars, training_epochs):
    """Learning rate schedule.

    :param se_hPars: Hyperparameters settings for layer.
    :type se_hPars: dict

    :param training_epochs: Number of training epochs for model.
    :type training_epochs: int

    :return: Updated settings for layer hyperparameters.
    :rtype: dict

    :return: Scheduled learning rate for layer.
    :rtype: list
    """
    # Extract hyperparameters
    e = se_hPars['epochs'] = training_epochs
    lr = se_hPars['learning_rate']
    d = se_hPars['cycle_descent']
    k = se_hPars['decay_k']

    # Compute dependent hyperparameters
    epc = se_hPars['cycle_epochs'] if se_hPars['cycle_epochs'] else training_epochs
    se_hPars['cycle_number'] = n = 1 if not epc else e // epc

    # Default decay
    if k == 0:
        # ~ 1% of initial lr for last epoch in cycle
        k = se_hPars['decay_k'] = 5 / epc

    # Compute learning rate schedule
    hPars = (e, lr, n, k, d, epc)
    lrate = schedule_functions(se_hPars['schedule'], hPars)

    return se_hPars, lrate



# EpyNN/epynn/dense/models.py
# Local application/library specific imports
from epynn.commons.models import Layer
from epynn.commons.maths import (
    sigmoid,
    xavier,
    activation_tune,
)
from epynn.dense.forward import dense_forward
from epynn.dense.backward import dense_backward
from epynn.dense.parameters import (
    dense_compute_shapes,
    dense_initialize_parameters,
    dense_compute_gradients,
    dense_update_parameters
)


class Dense(Layer):
    """
    Definition of a dense layer prototype.

    :param units: Number of units in dense layer, defaults to 1.
    :type units: int, optional

    :param activate: Non-linear activation of units, defaults to `sigmoid`.
    :type activate: function, optional

    :param initialization: Weight initialization function for dense layer, defaults to `xavier`.
    :type initialization: function, optional

    :param se_hPars: Layer hyper-parameters, defaults to `None` and inherits from model.
    :type se_hPars: dict[str, str or float] or NoneType, optional
    """

    def __init__(self,
                 units=1,
                 activate=sigmoid,
                 initialization=xavier,
                 se_hPars=None):
        """Initialize instance variable attributes.
        """
        super().__init__(se_hPars)

        self.d['u'] = units
        self.activate = activate
        self.initialization = initialization

        self.activation = { 'activate': self.activate.__name__ }
        self.trainable = True

        return None

    def compute_shapes(self, A):
        """Wrapper for :func:`epynn.dense.parameters.dense_compute_shapes()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`
        """
        dense_compute_shapes(self, A)

        return None

    def initialize_parameters(self):
        """Wrapper for :func:`epynn.dense.parameters.dense_initialize_parameters()`.
        """
        dense_initialize_parameters(self)

        return None

    def forward(self, A):
        """Wrapper for :func:`epynn.dense.forward.dense_forward()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`

        :return: Output of forward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        activation_tune(self.se_hPars)
        A = dense_forward(self, A)
        self.update_shapes(self.fc, self.fs)

        return A

    def backward(self, dX):
        """Wrapper for :func:`epynn.dense.backward.dense_backward()`.

        :param dX: Output of backward propagation from next layer.
        :type dX: :class:`numpy.ndarray`

        :return: Output of backward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        activation_tune(self.se_hPars)
        dX = dense_backward(self, dX)
        self.update_shapes(self.bc, self.bs)

        return dX

    def compute_gradients(self):
        """Wrapper for :func:`epynn.dense.parameters.dense_compute_gradients()`.
        """
        dense_compute_gradients(self)

        return None

    def update_parameters(self):
        """Wrapper for :func:`epynn.dense.parameters.dense_update_parameters()`.
        """
        if self.trainable:
            dense_update_parameters(self)

        return None



# EpyNN/epynnlive/music_author/train.py
# Standard library imports
import random

# Related third party imports
import numpy as np

# Local application/library specific imports
import epynn.initialize
from epynn.commons.maths import relu, softmax
from epynn.commons.library import (
    configure_directory,
    read_model,
)
from epynn.network.models import EpyNN
from epynn.embedding.models import Embedding
from epynn.rnn.models import RNN
from epynn.gru.models import GRU
from epynn.flatten.models import Flatten
from epynn.dropout.models import Dropout
from epynn.dense.models import Dense
from prepare_dataset import (
    prepare_dataset,
    download_music,
)
from settings import se_hPars


########################## CONFIGURE ##########################
random.seed(1)

np.set_printoptions(threshold=10)

np.seterr(all='warn')
np.seterr(under='ignore')

configure_directory()


############################ DATASET ##########################
download_music()

X_features, Y_label = prepare_dataset(N_SAMPLES=256)


####################### BUILD AND TRAIN MODEL #################
embedding = Embedding(X_data=X_features,
                      Y_data=Y_label,
                      X_encode=True,
                      Y_encode=True,
                      batch_size=16,
                      relative_size=(2, 1, 0))


### Feed-forward

# Model
name = 'Flatten_Dense-64-relu_Dense-2-softmax'

se_hPars['learning_rate'] = 0.01
se_hPars['softmax_temperature'] = 5

layers = [
    embedding,
    Flatten(),
    Dense(64, relu),
    Dropout(0.5),
    Dense(2, softmax),
]

model = EpyNN(layers=layers, name=name)

model.initialize(loss='MSE', seed=1, metrics=['accuracy', 'recall', 'precision'], se_hPars=se_hPars.copy())

model.train(epochs=10, init_logs=False)


### Recurrent

# Model
name = 'RNN-1-Seq_Flatten_Dense-64-relu_Dropout05_Dense-2-softmax'

se_hPars['learning_rate'] = 0.01
se_hPars['softmax_temperature'] = 1

layers = [
    embedding,
    RNN(1, sequences=True),
    Flatten(),
    Dense(64, relu),
    Dropout(0.5),
    Dense(2, softmax),
]

model = EpyNN(layers=layers, name=name)

model.initialize(loss='MSE', seed=1, metrics=['accuracy', 'recall', 'precision'], se_hPars=se_hPars.copy())

model.train(epochs=3, init_logs=False)


# Model
name = 'GRU-1-Seq_Flatten_Dense-64-relu_Dropout05_Dense-2-softmax'

se_hPars['learning_rate'] = 0.01
se_hPars['softmax_temperature'] = 1

layers = [
    embedding,
    GRU(1, sequences=True),
    Flatten(),
    Dense(64, relu),
    Dropout(0.5),
    Dense(2, softmax),
]

model = EpyNN(layers=layers, name=name)

model.initialize(loss='MSE', seed=1, metrics=['accuracy', 'recall', 'precision'], se_hPars=se_hPars.copy())

model.train(epochs=5, init_logs=False)


### Write/read model

model.write()
# model.write(path=/your/custom/path)

model = read_model()
# model = read_model(path=/your/custom/path)


### Predict

X_features, _ = prepare_dataset(N_SAMPLES=10)

dset = model.predict(X_features, X_encode=True)

for n, pred, probs in zip(dset.ids, dset.P, dset.A):
    print(n, pred, probs)



# EpyNN/epynn/embedding/backward.py


def initialize_backward(layer, dX):
    """Backward cache initialization.

    :param layer: An instance of embedding layer.
    :type layer: :class:`epynn.embedding.models.Embedding`

    :param dX: Output of backward propagation from next layer.
    :type dX: :class:`numpy.ndarray`

    :return: Input of backward propagation for current layer.
    :rtype: :class:`numpy.ndarray`
    """
    dA = layer.bc['dA'] = dX

    return dX


def embedding_backward(layer, dX):
    """Backward propagate error gradients to previous layer.
    """
    # (1) Initialize cache
    dA = initialize_backward(layer, dX)

    # (2) Pass backward
    dX = layer.bc['dX'] = dA

    return None    # No previous layer



# EpyNN/epynn/commons/models.py
# Related third party imports
import numpy as np
 

class Layer:
    """
    Definition of a parent **base layer** prototype. Any given **layer** prototype inherits from this class and is defined with respect to a specific architecture (Dense, RNN, Convolution...). The **parent** base layer defines instance attributes common to any **child** layer prototype.
    """

    def __init__(self, se_hPars=None):
        """Initialize instance variable attributes.

        :ivar d: Layer **dimensions** containing scalar quantities such as the number of nodes, hidden units, filters or samples.
        :vartype d: dict[str, int]

        :ivar fs: Layer **forward shapes** for parameters, input, output and processing intermediates.
        :vartype fs: dict[str, tuple[int]]

        :ivar p: Layer weight and bias **parameters**. These are the trainable parameters.
        :vartype p: dict[str, :class:`numpy.ndarray`]

        :ivar fc: Layer **forward cache** related for input, output and processing intermediates.
        :vartype fc: dict[str, :class:`numpy.ndarray`]

        :ivar bs: Layer **backward shapes** for gradients, input, output and processing intermediates.
        :vartype bs: dict[str, tuple[int]]

        :ivar g: Layer **gradients** used to update the trainable parameters.
        :vartype g: dict[str, :class:`numpy.ndarray`]

        :ivar bc: Layer **backward cache** for input, output and processing intermediates.
        :vartype bc: dict[str, :class:`numpy.ndarray`]

        :ivar o: Other scalar quantities that do not fit within the above-described attributes (rarely used).
        :vartype o: dict[str, int]

        :ivar activation: Conveniency attribute containing names of activation gates and corresponding activation functions.
        :vartype activation: dict[str, str]

        :ivar se_hPars: Layer **hyper-parameters**.
        :vartype se_hPars: dict[str, str or int]
        """
        self.d = {}
        self.fs = {}
        self.p = {}
        self.fc = {}
        self.bs = {}
        self.g = {}
        self.bc = {}
        self.o = {}

        self.activation = {}

        self.se_hPars = se_hPars

        return None

    def update_shapes(self, cache, shapes):
        """Update shapes from cache.

        :param cache: Cache from forward or backward propagation.
        :type cache: dict[str, :class:`numpy.ndarray`]

        :param shapes: Corresponding shapes.
        :type shapes: dict[str, tuple[int]]
        """
        shapes.update({k:v.shape for k,v in cache.items()})

        return None


class dataSet:
    """
    Definition of a dataSet object prototype.

    :param X_data: Set of sample features.
    :type X_data: list[list[int or float]]

    :param Y_data: Set of sample label, defaults to None.
    :type Y_data: list[list[int] or int] or NoneType, optional

    :param name: Name of set, defaults to 'dummy'.
    :type name: str, optional
    """

    def __init__(self,
                 X_data,
                 Y_data=None,
                 name='dummy'):
        """Initialize dataSet object.

        :ivar active: `True` if X_data is not empty.
        :vartype active: bool

        :ivar X: Set of sample features.
        :vartype X: :class:`numpy.ndarray`

        :ivar Y: Set of sample label.
        :vartype Y: :class:`numpy.ndarray`

        :ivar y: Set of single-digit sample label.
        :vartype y: :class:`numpy.ndarray`

        :ivar b: Balance of labels in set.
        :vartype b: dict[int: int]

        :ivar ids: Sample identifiers.
        :vartype ids: list[int]

        :ivar A: Output of forward propagation.
        :vartype A: :class:`numpy.ndarray`

        :ivar P: Label predictions.
        :vartype P: :class:`numpy.ndarray`

        :ivar name: Name of dataset.
        :vartype name: str
        """
        self.name = name

        self.active = True if len(X_data) > 0 else False

        # Set of sample features
        if self.active:

            # Vectorize X_data in NumPy array
            self.X = np.array(X_data)

        # Set of sample label
        if (hasattr(Y_data, 'shape') or Y_data) and self.active:

            # Vectorize Y_data in NumPy array
            Y_data = np.array(Y_data)

            # Check if labels are one-hot encoded or not
            is_encoded = True if Y_data.ndim == 2 else False

            # If not encoded, reshape from (N_SAMPLES,) to (N_SAMPLES, 1)
            self.Y = Y_data if is_encoded else np.expand_dims(Y_data, 1)

            # Retrieve single-digit label w. r. t. one-hot encoding
            self.y = np.argmax(Y_data, axis=1) if is_encoded else Y_data

            # Map single-digit label with representation in dataset
            self.b = {label:np.count_nonzero(self.y == label) for label in self.y}

        # Set sample identifiers
        self.ids = np.array([i for i in range(len(X_data))])

        # Initialize empty arrays
        self.A = np.array([])
        self.P = np.array([])

        return None



# EpyNN/epynn/dense/parameters.py
# Related third party imports
import numpy as np


def dense_compute_shapes(layer, A):
    """Compute forward shapes and dimensions from input for layer.
    """
    X = A    # Input of current layer

    layer.fs['X'] = X.shape    # (m, n)

    layer.d['m'] = layer.fs['X'][0]    # Number of samples  (m)
    layer.d['n'] = layer.fs['X'][1]    # Number of features (n)

    # Shapes for trainable parameters              Units (u)
    layer.fs['W'] = (layer.d['n'], layer.d['u'])    # (n, u)
    layer.fs['b'] = (1, layer.d['u'])               # (1, u)

    return None


def dense_initialize_parameters(layer):
    """Initialize trainable parameters from shapes for layer.
    """
    # For linear activation of inputs (Z)
    layer.p['W'] = layer.initialization(layer.fs['W'], rng=layer.np_rng)
    layer.p['b'] = np.zeros(layer.fs['b']) # Z = dot(X, W) + b

    return None


def dense_compute_gradients(layer):
    """Compute gradients with respect to weight and bias for layer.
    """
    X = layer.fc['X']      # Input of forward propagation
    dZ = layer.bc['dZ']    # Gradient of the loss with respect to Z

    # (1) Gradient of the loss with respect to W, b
    dW = layer.g['dW'] = np.dot(X.T, dZ)       # (1.1) dL/dW
    db = layer.g['db'] = np.sum(dZ, axis=0)    # (1.2) dL/db

    return None


def dense_update_parameters(layer):
    """Update parameters from gradients for layer.
    """
    for gradient in layer.g.keys():
        parameter = gradient[1:]
        # Update is driven by learning rate and gradients
        layer.p[parameter] -= layer.lrate[layer.e] * layer.g[gradient]

    return None



# EpyNN/epynnlive/ptm_protein/prepare_dataset.py
# Standard library imports
import tarfile
import random
import os

# Related third party imports
import wget

# Local application/library specific imports
from epynn.commons.library import read_file
from epynn.commons.logs import process_logs


def download_sequences():
    """Download a set of peptide sequences.
    """
    data_path = os.path.join('.', 'data')

    if not os.path.exists(data_path):

        # Download @url with wget
        url = 'https://synthase.s3.us-west-2.amazonaws.com/ptm_prediction_data.tar'
        fname = wget.download(url)

        # Extract archive
        tar = tarfile.open(fname).extractall('.')
        process_logs('Make: ' + fname, level=1)

        # Clean-up
        os.remove(fname)

    return None


def prepare_dataset(N_SAMPLES=100):
    """Prepare a set of labeled peptides.

    :param N_SAMPLES: Number of peptide samples to retrieve, defaults to 100.
    :type N_SAMPLES: int

    :return: Set of peptides.
    :rtype: tuple[list[str]]

    :return: Set of single-digit peptides label.
    :rtype: tuple[int]
    """
    # Single-digit positive and negative labels
    p_label = 0
    n_label = 1

    # Positive data are Homo sapiens O-GlcNAcylated peptide sequences from oglcnac.mcw.edu
    path_positive = os.path.join('data', '21_positive.dat')

    # Negative data are peptide sequences presumably not O-GlcNAcylated
    path_negative = os.path.join('data', '21_negative.dat')

    # Read text files, each containing one sequence per line
    positive = [[list(x), p_label] for x in read_file(path_positive).splitlines()]
    negative = [[list(x), n_label] for x in read_file(path_negative).splitlines()]

    # Shuffle data to prevent from any sorting previously applied
    random.shuffle(positive)
    random.shuffle(negative)

    # Truncate to prepare a balanced dataset
    negative = negative[:len(positive)]

    # Prepare a balanced dataset
    dataset = positive + negative

    # Shuffle dataset
    random.shuffle(dataset)

    # Truncate dataset to N_SAMPLES
    dataset = dataset[:N_SAMPLES]

    # Separate X-Y pairs
    X_features, Y_label = zip(*dataset)

    return X_features, Y_label



# EpyNN/epynn/convolution/models.py
# Local application/library specific imports
from epynn.commons.models import Layer
from epynn.commons.maths import (
    relu,
    xavier,
    activation_tune,
)
from epynn.convolution.forward import convolution_forward
from epynn.convolution.backward import convolution_backward
from epynn.convolution.parameters import (
    convolution_compute_shapes,
    convolution_initialize_parameters,
    convolution_compute_gradients,
    convolution_update_parameters,
)


class Convolution(Layer):
    """
    Definition of a convolution layer prototype.

    :param unit_filters: Number of unit filters in convolution layer, defaults to 1.
    :type unit_filters: int, optional

    :param filter_size: Height and width for convolution window, defaults to `(3, 3)`.
    :type filter_size: int or tuple[int], optional

    :param strides: Height and width to shift the convolution window by, defaults to `None` which equals `filter_size`.
    :type strides: int or tuple[int], optional

    :param padding: Number of zeros to pad each features plane with, defaults to 0.
    :type padding: int, optional

    :param activate: Non-linear activation of unit filters, defaults to `relu`.
    :type activate: function, optional

    :param initialization: Weight initialization function for convolution layer, defaults to `xavier`.
    :type initialization: function, optional

    :param use_bias: Whether the layer uses bias, defaults to `True`.
    :type use_bias: bool, optional

    :param se_hPars: Layer hyper-parameters, defaults to `None` and inherits from model.
    :type se_hPars: dict[str, str or float] or NoneType, optional
    """

    def __init__(self,
                 unit_filters=1,
                 filter_size=(3, 3),
                 strides=None,
                 padding=0,
                 activate=relu,
                 initialization=xavier,
                 use_bias=True,
                 se_hPars=None):
        """Initialize instance variable attributes.
        """
        super().__init__()

        filter_size = filter_size if isinstance(filter_size, tuple) else (filter_size, filter_size)
        strides = strides if isinstance(strides, tuple) else filter_size

        self.d['u'] = unit_filters
        self.d['fh'], self.d['fw'] = filter_size
        self.d['sh'], self.d['sw'] = strides
        self.d['p'] = padding
        self.activate = activate
        self.initialization = initialization
        self.use_bias = use_bias

        self.activation = { 'activate': activate.__name__ }
        self.trainable = True

        return None

    def compute_shapes(self, A):
        """Wrapper for :func:`epynn.convolution.parameters.convolution_compute_shapes()`.

        :param A: Output of forward propagation from previous layer.
        :type A: :class:`numpy.ndarray`
        """
        convolution_compute_shapes(self, A)

        return None

    def initialize_parameters(self):
        """Wrapper for :func:`epynn.convolution.parameters.convolution_initialize_parameters()`.
        """
        convolution_initialize_parameters(self)

        return None

    def forward(self, A):
        """Wrapper for :func:`epynn.convolution.forward.convolution_forward()`.

        :param A: Output of forward propagation from *previous* layer.
        :type A: :class:`numpy.ndarray`

        :return: Output of forward propagation for **current** layer.
        :rtype: :class:`numpy.ndarray`
        """
        activation_tune(self.se_hPars)
        A = convolution_forward(self, A)
        self.update_shapes(self.fc, self.fs)

        return A

    def backward(self, dX):
        """Wrapper for :func:`epynn.convolution.backward.convolution_backward()`.

        :param dX: Output of backward propagation from next layer.
        :type dX: :class:`numpy.ndarray`

        :return: Output of backward propagation for current layer.
        :rtype: :class:`numpy.ndarray`
        """
        activation_tune(self.se_hPars)
        dX = convolution_backward(self, dX)
        self.update_shapes(self.bc, self.bs)

        return dX

    def compute_gradients(self):
        """Wrapper for :func:`epynn.convolution.parameters.convolution_compute_gradients()`.
        """
        convolution_compute_gradients(self)

        return None

    def update_parameters(self):
        """Wrapper for :func:`epynn.convolution.parameters.convolution_update_parameters()`.
        """
        if self.trainable:
            convolution_update_parameters(self)

        return None
